"""State module for managing Amazon IAM Policies."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    policy_document: Dict or str,
    resource_id: str = None,
    path: str = "/",
    description: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    timeout: make_dataclass(
        """Timeout configuration for AWS IAM Policy.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    """Timeout configuration when creating an AWS IAM Policy."""
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=1)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    """Timeout configuration when updating an AWS IAM Policy."""
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=1)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates or updates an AWS IAM Policy.

    When creating a policy, this operation creates a policy version with a version identifier of `v1` and sets `v1` as the
    policy's default version. When updating a policy, this operation creates a new policy version, sets the new policy to
    the default version, and deletes the old policy.

    Args:
        name(str):
            The name of the IAM Policy.
        policy_document(dict or str):
            The JSON policy document that you want to use as the content for the new policy.
            You must provide policies in JSON format in IAM. However, for CloudFormation templates formatted in YAML,
            you can provide the policy in JSON or YAML format. CloudFormation always converts a YAML policy to JSON format
            before submitting it to IAM.
        resource_id(str, Optional):
            The Amazon Resource Name (ARN) of the IAM policy in Amazon Web Services.
        path(str, Optional):
            The path for the policy. This parameter is optional. If it is not included, it defaults to a slash (``/``).
        description(str, Optional):
            A friendly description of the policy.
        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the policy.

            * Key (*str*):
                The key of the tag.
            * Value (*str*):
                The value of the tag.
        timeout(dict, Optional):
            Timeout configuration for AWS IAM Policy.

            * create (*dict, Optional*):
                Timeout configuration when creating an AWS IAM Policy.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``1``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``40``.

            * update (*dict, Optional*):
                Timeout configuration when updating an AWS IAM Policy.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``1``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``40``.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_iam_policy]:
          aws.iam.policy.present:
            - name: 'string'
            - resource_id: 'string'
            - policy_document: 'dict' or 'string'
            - path: 'string'
            - description: 'string'
            - tags:
                - Key: 'string'
                  Value: 'string'
            - timeout:
                create:
                  delay: int
                  max_attemps: int
                update:
                  delay: int
                  max_attemps: int

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_iam_policy:
              aws.iam.policy.present:
                - name: 'idem_test_iam_policy'
                - policy_document:
                    Version: '2012-10-17'
                    Statement:
                      - Sid: 'AllowCreateSubnet'
                        Effect: 'Allow'
                        Action: ['ec2:CreateSubnet']
                        Resource: '*'
                - path: '/'
                - description: 'My IAM Policy'
                - tags:
                    - Key: 'provider'
                      Value: 'idem'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    update_ret = None

    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )

    # Standardise on the json format
    policy_document = hub.tool.aws.state_comparison_utils.standardise_json(
        policy_document
    )

    if resource_id:
        before = await hub.exec.aws.iam.policy.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = before["ret"]
        plan_state = copy.deepcopy(result["old_state"])
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.iam.policy", name=name
        )
        # Update policy document if needed
        if (
            result["old_state"] is not None
            and result["old_state"]["policy_document"]
            and not hub.tool.aws.state_comparison_utils.is_json_identical(
                result["old_state"]["policy_document"], policy_document
            )
        ):
            if ctx.get("test", False):
                plan_state["policy_document"] = policy_document
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.iam.policy", name=name
                )
            else:
                update_ret = await hub.tool.aws.iam.policy.update(
                    ctx=ctx,
                    policy_arn=resource_id,
                    policy_version_id=result["old_state"].get("default_version_id"),
                    new_policy_document=policy_document,
                    timeout=timeout,
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = False

        if (
            result["old_state"] is not None
            and tags is not None
            and tags != result["old_state"].get("tags")
        ):
            update_ret = await hub.tool.aws.iam.policy.update_tags(
                ctx=ctx,
                police_arn=resource_id,
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            if not update_ret["result"]:
                result["result"] = False
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]

        if update_ret is not None and result["result"] and not ctx.get("test", False):
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.iam.policy", name=name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": f"policy-{name}-resource_id",
                    "id": f"policy-{name}-id",
                    "path": path,
                    "policy_document": policy_document,
                    "description": description,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.iam.policy", name=name
            )
            return result

        ret = await hub.exec.boto3.client.iam.create_policy(
            ctx,
            PolicyName=name,
            Path=path,
            PolicyDocument=policy_document,
            Description=description,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            if tags
            else None,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=1,
            default_max_attempts=40,
            timeout_config=timeout.get("create") if timeout else None,
        )
        hub.log.debug(f"Waiting on creation of aws.iam.policy '{name}'")
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "iam",
                "policy_exists",
                PolicyArn=ret["ret"].get("Policy")["Arn"],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.iam.policy", name=name
        )
        resource_id = ret["ret"]["Policy"]["Arn"]

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif before and (update_ret is None):
            result["new_state"] = copy.deepcopy(result["old_state"])
        else:
            after = await hub.exec.aws.iam.policy.get(
                ctx, name=name, resource_id=resource_id
            )
            result["new_state"] = after["ret"]
    except Exception as e:
        result["comment"] = (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the specified AWS IAM Policy.

    Before you can delete a managed policy, you must first detach the policy from all users, groups, and roles that it is attached to.
    In addition, you must delete all the policy's versions. The following steps describe the process for deleting a managed policy:

    * Detach the policy from all users, groups, and roles that the policy is attached to.
    * Delete all versions of the policy.
    * Delete the policy (this automatically deletes the policy's default version).

    Args:
        name(str):
            The name of the IAM Policy.
        resource_id(str, Optional):
            The Amazon Resource Name (ARN) of the IAM policy in Amazon Web Services.

            .. warning::
              Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_iam_policy]:
          aws.iam.policy.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_iam_policy:
              aws.iam.policy:
                - name: 'idem_test_iam_policy'
                - resource_id: 'arn:aws:iam::123456789012:policy/idem_test_iam_policy'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.policy", name=name
        )
        return result

    before = await hub.exec.aws.iam.policy.get(ctx, name=name, resource_id=resource_id)
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.policy", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.iam.policy", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.iam.delete_policy(ctx, PolicyArn=resource_id)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.iam.policy", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes IAM Local Policies in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.iam.policy
    """
    result = {}

    # Set scope to local to only list the customer-defined managed policies.
    ret = await hub.exec.aws.iam.policy.list(ctx, scope="Local")
    if not ret["result"]:
        hub.log.debug(f"Could not describe policies {ret['comment']}")
        return {}

    for policy in ret["ret"]:
        resource_key = f"iam-policy-{policy['resource_id']}"
        result[resource_key] = {
            "aws.iam.policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in policy.items()
            ]
        }

    return result



"""

Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_security_group
hub.exec.boto3.client.ec2.delete_security_group
hub.exec.boto3.client.ec2.describe_security_groups
hub.tool.boto3.resource.exec(resource, authorize_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, authorize_ingress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_ingress, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.security_group.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    group_id: str,
    ip_protocol: str,
    from_port: int,
    to_port: int,
    is_egress: bool,
    resource_id: str = None,
    cidr_ipv4: str = None,
    cidr_ipv6: str = None,
    prefix_list_id: str = None,
    referenced_group_info: Dict = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    description: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a security group rule.
    authorize_security_group_ingress: Adds the specified inbound (ingress) rules to a security group. An inbound rule
    permits instances to receive traffic from the specified IPv4 or IPv6 CIDR address range, or from the instances that
    are associated with the specified destination security groups.You specify a protocol for each rule (for example, TCP).
    For TCP and UDP, you must also specify the destination port or port range. For ICMP/ICMPv6, you must also specify the
    ICMP/ICMPv6 type and code. You can use -1 to mean all types or all codes. Rule changes are propagated to instances within
    the security group as quickly as possible. However, a small delay might occur.
    authorize_security_group_egress: [VPC only] Adds the specified outbound (egress) rules to a security group for use with a VPC.
    An outbound rule permits instances to send traffic to the specified IPv4 or IPv6 CIDR address ranges, or to the instances that
    are associated with the specified source security groups. You specify a protocol for each rule (for example, TCP). For the TCP
    and UDP protocols, you must also specify the destination port or port range. For the ICMP protocol, you must also specify the
    ICMP type and code. You can use -1 for the type or code to mean all types or all codes. Rule changes are propagated to affected
    instances as quickly as possible. However, a small delay might occur.

    Args:
        name(Text): An Idem name to identify the security group rule resource.
        group_id(str): The ID of the security group
        is_egress(bool): To find the type of rule, whether it is a ingress or egress.
        ip_protocol(str): The IP protocol name (tcp , udp , icmp , icmpv6 ) or number (see Protocol Numbers ). [VPC only] Use -1 to specify all protocols.
                          When authorizing security group rules, specifying -1 or a protocol number other than tcp , udp , icmp , or icmpv6 allows traffic
                          on all ports, regardless of any port range you specify. For tcp , udp , and icmp , you must specify a port range. For icmpv6 ,
                          the port range is optional; if you omit the port range, traffic for all types and codes is allowed.
        from_port(int): The start of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 type number. A value of -1 indicates all ICMP/ICMPv6 types.
                        If you specify all ICMP/ICMPv6 types, you must specify all codes.
        to_port(int): The end of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 code. A value of -1 indicates all ICMP/ICMPv6 codes.
                      If you specify all ICMP/ICMPv6 types, you must specify all codes.
        cidr_ipv4(str, optional): The IPv4 CIDR range. You can either specify a CIDR range or a source security group, not both.
                        To specify a single IPv4 address, use the /32 prefix length.
        resource_id(Text, optional): AWS Security group rule ID.
        cidr_ipv6(str, optional): The IPv6 CIDR range. You can either specify a CIDR range or a source security group, not both.
                                  To specify a single IPv6 address, use the /128 prefix length.
        prefix_list_id(str, optional): The ID of the prefix.
        referenced_group_info(Dict, optional): The security group and Amazon Web Services account ID pairs.
        description(str, optional): The description of the security group rule.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the security group rule.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
    Request Syntax:
        [security_group_rule_name]:
          aws.ec2.security_group_rule.present:
            - group_id: 'string'
            - is_egress: 'bool'
            - ip_protocol: 'string'
            - from_port: 'int'
            - to_port: 'int'
            - cidr_ipv4: 'string'
            - cidr_ipv6: 'string'
            - prefix_list_id: 'string'
            - referenced_group_info: 'dict'
            - tags:
                - Key: 'string'
                  Value: 'string'
            - description: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-sg-rule:
              aws.ec2.security_group_rule.present:
                - group_id: sg-0dd442ba9f89c5d59
                - is_egress: false
                - ip_protocol: tcp
                - from_port: 60
                - to_port: 60
                - cidr_ipv4: 0.0.0.0/0
                - tags:
                    - Key: test_name5
                      Value: test-rule5
                - description: 'Security group desc'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.ec2.describe_security_group_rules(
            ctx, SecurityGroupRuleIds=[resource_id]
        )
    if before and before["result"] and before["ret"]["SecurityGroupRules"]:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
            raw_resource=before["ret"]["SecurityGroupRules"][0], idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = (
            await hub.exec.aws.ec2.security_group_rule.update_security_group_rule(
                ctx,
                before=before["ret"]["SecurityGroupRules"][0],
                cidr_ipv4=cidr_ipv4,
                cidr_ipv6=cidr_ipv6,
                description=description,
                from_port=from_port,
                to_port=to_port,
                group_id=group_id,
                ip_protocol=ip_protocol,
                prefix_list_id=prefix_list_id,
                referenced_group_info=referenced_group_info,
                resource_id=resource_id,
            )
        )

        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            for key, value in update_ret["ret"].items():
                plan_state[key] = value
        if not resource_updated:
            result["comment"] = result["comment"] + (
                f"aws.ec2.security_group_rule {name} already exists",
            )
    else:
        if isinstance(tags, List):
            tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "group_id": group_id,
                    "ip_protocol": ip_protocol,
                    "from_port": from_port,
                    "to_port": to_port,
                    "is_egress": is_egress,
                    "cidr_ipv4": cidr_ipv4,
                    "tags": tags,
                    "description": description,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.security_group_rule", name=name
            )
            return result
        try:
            ip_permissions = []
            ip_perm = {}
            if ip_protocol:
                ip_perm["IpProtocol"] = ip_protocol
            if from_port:
                ip_perm["FromPort"] = from_port
            if to_port:
                ip_perm["ToPort"] = to_port
            if cidr_ipv4:
                IpRanges = {}
                if description:
                    IpRanges["Description"] = description
                if cidr_ipv4:
                    IpRanges["CidrIp"] = cidr_ipv4
                ip_perm["IpRanges"] = [IpRanges]
            if cidr_ipv6:
                Ipv6Ranges = {}
                if description:
                    Ipv6Ranges["Description"] = description
                if cidr_ipv6:
                    Ipv6Ranges["CidrIpv6"] = cidr_ipv6
                ip_perm["Ipv6Ranges"] = [Ipv6Ranges]
            if prefix_list_id:
                prefix_list = {}
                if description:
                    prefix_list["Description"] = description
                if prefix_list_id:
                    prefix_list["PrefixListId"] = prefix_list_id
                ip_perm["PrefixListIds"] = [prefix_list]
            if referenced_group_info:
                if description:
                    referenced_group_info["Description"] = description
                ip_perm["UserIdGroupPairs"] = [referenced_group_info]
            ip_permissions.append(ip_perm)
            if not is_egress:
                ret = await hub.exec.boto3.client.ec2.authorize_security_group_ingress(
                    ctx,
                    GroupId=group_id,
                    IpPermissions=ip_permissions,
                    TagSpecifications=[
                        {
                            "ResourceType": "security-group-rule",
                            "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                                tags
                            ),
                        }
                    ]
                    if tags
                    else None,
                )
                result["result"] = ret["result"]
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.ec2.security_group_rule", name=name
                )
            else:
                ret = await hub.exec.boto3.client.ec2.authorize_security_group_egress(
                    ctx,
                    GroupId=group_id,
                    IpPermissions=ip_permissions,
                    TagSpecifications=[
                        {
                            "ResourceType": "security-group-rule",
                            "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                                tags
                            ),
                        }
                    ]
                    if tags
                    else None,
                )
                result["result"] = ret["result"]
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.ec2.security_group_rule", name=name
                )
            resource_id = ret["ret"]["SecurityGroupRules"][0]["SecurityGroupRuleId"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.ec2.describe_security_group_rules(
                ctx,
                SecurityGroupRuleIds=[resource_id],
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
                raw_resource=after["ret"]["SecurityGroupRules"][0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    group_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a security group rule.

    Args:
        name(Text): An Idem name to identify the security group rule resource.
        group_id(Text, optional): AWS Security Group ID. Idem automatically considers this resource being absent
         if this field is not specified.
        resource_id(Text, optional): AWS Security Group rule ID. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.security_group_rule.absent:
                - name: value
                - group_id: "sg-0008bd25b7867b5cf"
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if (not resource_id) or (not group_id):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.security_group_rule", name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "SecurityGroup", group_id)
    before = await hub.exec.boto3.client.ec2.describe_security_group_rules(
        ctx, SecurityGroupRuleIds=[resource_id]
    )
    if not before and not before["result"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.security_group_rule", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
            raw_resource=before["ret"]["SecurityGroupRules"][0], idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ec2.security_group_rule", name=name
            )
            return result
        try:
            # checks for egress or ingress using before's output and revoke function needed sg resource
            if not before["ret"]["SecurityGroupRules"][0]["IsEgress"]:
                ret = resource.revoke_ingress(SecurityGroupRuleIds=[resource_id])
                result["result"] = ret["Return"]
                if not result["result"]:
                    result["comment"] = (
                        f"aws.ec2.security_group_rule '{name}' is already in deleted state.",
                    )
                    return result
                result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.ec2.security_group_rule", name=name
                )
            else:
                ret = resource.revoke_egress(SecurityGroupRuleIds=[resource_id])
                result["result"] = ret["Return"]
                if not result["result"]:
                    result["comment"] = (
                        f"aws.ec2.security_group_rule '{name}' is already in deleted state.",
                    )
                    return result
                result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.ec2.security_group_rule", name=name
                )
        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.security_group_rule
    """
    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_security_group_rules(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe security_group rule {ret['comment']}")
        result["comment"] = ret["comment"]
        result["result"] = False
        return result

    for resource in ret["ret"]["SecurityGroupRules"]:
        resource_id = resource.get("SecurityGroupRuleId")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
                raw_resource=resource, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.security_group_rule.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS Lambda Function Permission resources.
"""


def convert_raw_lambda_permission_to_present(
    hub,
    raw_resource: Dict[str, Any],
    function_name: str,
    revision_id: str = None,
) -> Dict[str, Any]:
    r"""
    Converts an AWS Lambda Function Permission resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        function_name(string): The name of the AWS Lambda Function.
        revision_id(string, optional): The revision ID of the policy.

    Returns:
        Dict[str, Any]: Common idem present state.
    """
    describe_parameters = OrderedDict(
        {
            "Sid": "resource_id",
            "Action": "action",
            "Principal": "principal",
            "Effect": "effect",
        }
    )
    translated_resource = {}
    for parameter_raw, parameter_present in describe_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)
    translated_resource["name"] = raw_resource["Sid"]
    translated_resource["function_name"] = function_name

    if revision_id:
        translated_resource["revision_id"] = revision_id

    qualifier = hub.tool.aws.arn_utils.get_qualifier(raw_resource.get("Resource"))
    if qualifier:
        translated_resource["qualifier"] = qualifier

    condition = raw_resource.get("Condition")
    if condition:
        stringEquals = condition.get("StringEquals")
        if stringEquals:
            source_account = stringEquals.get("AWS:SourceAccount")
            if source_account:
                translated_resource["source_account"] = source_account
            event_source_token = stringEquals.get("lambda:EventSourceToken")
            if event_source_token:
                translated_resource["event_source_token"] = event_source_token
            principal_org_id = stringEquals.get("aws:PrincipalOrgID")
            if principal_org_id:
                translated_resource["principal_org_id"] = principal_org_id
            function_url_auth_type = stringEquals.get("lambda:FunctionUrlAuthType")
            if function_url_auth_type:
                translated_resource["function_url_auth_type"] = function_url_auth_type

        arnLike = condition.get("ArnLike")
        if arnLike:
            translated_resource["source_arn"] = arnLike.get("AWS:SourceArn")

    return translated_resource



"""State module for managing Amazon Config Recorder."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    role_arn: str,
    recording_group: make_dataclass(
        "ConfigurationRecorder",
        [
            ("allSupported", bool, field(default=None)),
            ("includeGlobalResourceTypes", bool, field(default=None)),
            ("resourceTypes", List[str], field(default=None)),
        ],
    ),
    recording: bool = False,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Creates a new configuration recorder to record the selected resource configurations.

    Args:
        name(str):
            The name of the recorder.

        role_arn (str):
            Amazon Resource Name (ARN) of the IAM role used to describe the Amazon Web Services resources
            associated with the account.

        recording_group (dict[str, Any], Optional):
            Specifies the types of Amazon Web Services resources for which Config records configuration changes.

            * allSupported (bool, Optional):
              Specifies whether Config records configuration changes for every supported type of regional resource.

              If you set this option to true, when Config adds support for a new type of regional resource,
              it starts recording resources of that type automatically.

              If you set this option to true, you cannot enumerate a list of resourceTypes.

            * includeGlobalResourceTypes (bool, Optional):
              Specifies whether Config includes all supported types of global resources (for example, IAM
              resources) with the resources that it records.

              Before you can set this option to true, you must set the allSupported option to true.

              If you set this option to true, when Config adds support for a new type of global resource,
              it starts recording resources of that type automatically.

              The configuration details for any global resource are the same in all regions. To prevent duplicate
              configuration items, you should consider customizing Config in only one region to record global
              resources.

            * resourceTypes (list[str], Optional):
              A comma-separated list that specifies the types of Amazon Web Services resources for which
              Config records configuration changes (for example, AWS::EC2::Instance or AWS::CloudTrail::Trail).

              To record all configuration changes, you must set the allSupported option to true.

              If you set this option to false, when Config adds support for a new type of resource, it will not record
              resources of that type unless you manually add that type to your recording group.

              For a list of valid resourceTypes values, see the resourceType Value column in Supported
              Amazon Web Services resource Types.

        recording (bool, Optional):
            Specifies recording status of the configuration recorder. Default is False.

        resource_id (str, Optional):
            The name of the recorder.

    Request syntax:
        .. code-block:: sls

            [aws-config-recorder]:
              aws.config.config_recorder.present:
              - name: 'string'
              - resource_id: 'string'
              - role_arn: 'string'
              - recording: 'string'
              - recording_group: 'dict'

    Returns:
         Dict[str, Any]

    Examples:
        .. code-block:: sls

             aws-config-recorder:
               aws.config.config_recorder.present:
                 - name: 'config_recorder'
                 - resource_id: 'config_recorder'
                 - role_arn: 'arn:aws:iam::012345678912:role/aws-service-role/config.amazonaws.com/AWSServiceRoleForConfig'
                 - recording: true
                 - recording_group:
                    allSupported: false
                    includeGlobalResourceTypes: false
                    resourceTypes:
                    - "AWS::ApiGateway::Stage"
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated: bool = False
    only_recording_status_changed: bool = False

    if resource_id:
        resource = await hub.exec.boto3.client.config.describe_configuration_recorders(
            ctx,
            ConfigurationRecorderNames=[name],
        )
        if not resource["result"]:
            result["comment"] = resource["comment"]
            result["result"] = False
            return result
        else:
            before = resource["ret"]["ConfigurationRecorders"][0]

    if before:
        before_recording_status = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not before_recording_status["result"]:
            result["comment"] = before_recording_status["comment"]
            result["result"] = False
            return result
        before["recording"] = before_recording_status["ret"][
            "ConfigurationRecordersStatus"
        ][0]["recording"]

        resource_translated = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=before,
            recording=before_recording_status["ret"]["ConfigurationRecordersStatus"][0][
                "recording"
            ],
        )
        result["old_state"] = resource_translated
        resource_updated = await hub.tool.aws.config.config_utils.is_resource_updated(
            before=resource_translated,
            role_arn=role_arn,
            recording_group=recording_group,
        )
        if not resource_updated and resource_translated["recording"] == recording:
            result["comment"] = (f"aws.config.config_recorder '{name}' already exists",)
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        if not resource_updated and resource_translated["recording"] != recording:
            only_recording_status_changed = True

    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={},
            desired_state={
                "name": name,
                "role_arn": role_arn,
                "recording": recording,
                "resource_id": resource_id,
                "recording_group": recording_group,
            },
        )
        operation = (
            "update"
            if before
            and (resource_updated or resource_translated["recording"] != recording)
            else "create"
        )
        result["comment"] = (f"Would {operation} aws.config.config_recorder '{name}'",)
        return result

    if not only_recording_status_changed:
        update_ret = await hub.exec.boto3.client.config.put_configuration_recorder(
            ctx=ctx,
            ConfigurationRecorder={
                "name": name,
                "roleARN": role_arn,
                "recordingGroup": recording_group,
            },
        )
        if not update_ret["result"]:
            result["comment"] = update_ret["comment"]
            result["result"] = False
            return result

        after_ret = await hub.exec.boto3.client.config.describe_configuration_recorders(
            ctx,
            ConfigurationRecorderNames=[name],
        )
        if not after_ret["result"]:
            result["comment"] = after_ret["comment"]
            result["result"] = False
            return result

        result[
            "new_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=after_ret["ret"]["ConfigurationRecorders"][0],
            recording=None,
        )
        operation = "Created" if not before else "Updated"
        result["comment"] = (f"{operation} aws.config.config_recorder '{name}'",)
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    # If it is create recorder case and recording is True
    #   OR
    # If it is update recorder case and recording status is changed
    if (not before and recording) or (before and recording != before["recording"]):
        if recording:
            recording_status_ret = (
                await hub.exec.boto3.client.config.start_configuration_recorder(
                    ctx=ctx, ConfigurationRecorderName=name
                )
            )
        else:
            recording_status_ret = (
                await hub.exec.boto3.client.config.stop_configuration_recorder(
                    ctx=ctx, ConfigurationRecorderName=name
                )
            )
        if not recording_status_ret["result"]:
            result["comment"] = recording_status_ret["comment"]
            result["result"] = False
            return result

        after_recording_status_ret = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not after_recording_status_ret["result"]:
            result["comment"] = after_recording_status_ret["comment"]
            result["result"] = False
            return result
        result["new_state"]["recording"] = after_recording_status_ret["ret"][
            "ConfigurationRecordersStatus"
        ][0]["recording"]
        if only_recording_status_changed:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.config.config_recorder", name=name
            )
    else:
        if not before:
            result["new_state"]["recording"] = False
        else:
            result["new_state"]["recording"] = before["recording"]
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the configuration recorder.

    After the configuration recorder is deleted, Config will not record resource configuration changes until you
    create a new configuration recorder.

    Args:
        name(str):
            The name of the recorder.

        resource_id(str, Optional):
            AWS Config configuration recorder Name. Idem automatically considers this resource being absent
            if this field is not specified.

    Returns:
          Dict[str, Any]

    Examples:
          .. code-block:: sls

            aws-config-recorder:
              aws.config.config_recorder.absent:
                - name: 'config_recorder'
                - resource_id: 'config_recorder'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        resource = await hub.exec.boto3.client.config.describe_configuration_recorders(
            ctx,
            ConfigurationRecorderNames=[name],
        )
        if not resource["result"]:
            if "NoSuchConfigurationRecorderException" in str(resource["comment"]):
                result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                    resource_type="aws.config.config_recorder", name=name
                )
                return result
            result["comment"] = resource["comment"]
            result["result"] = False
            return result
        else:
            before = resource["ret"]["ConfigurationRecorders"][0]

    if before:
        before_recording = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not before_recording["result"]:
            result["comment"] = before_recording["comment"]
            result["result"] = False
            return result

        translated_resource = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=before,
            recording=before_recording["ret"]["ConfigurationRecordersStatus"][0][
                "recording"
            ],
        )
        result["old_state"] = translated_resource
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.config_recorder", name=name
        )
        return result

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.config.config_recorder", name=name
        )
        return result
    else:
        delete_ret = await hub.exec.boto3.client.config.delete_configuration_recorder(
            ctx, ConfigurationRecorderName=name
        )
        result["result"] = delete_ret["result"]
        if not result["result"]:
            result["comment"] = delete_ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.config.config_recorder", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Return details about your config recorder.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.config.config_recorder
    """
    result = {}
    ret = await hub.exec.boto3.client.config.describe_configuration_recorders(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe configuration recorder {ret['comment']}")
        return {}

    for resource in ret["ret"]["ConfigurationRecorders"]:
        name = resource["name"]
        recording_status = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not recording_status["result"]:
            hub.log.warning(
                f"Could not fetch configuration recorder '{name}' status with error {recording_status['comment']}"
            )
            continue

        recorder_name = f"{name}-config-recorder"
        translated_resource = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=resource,
            recording=recording_status["ret"]["ConfigurationRecordersStatus"][0][
                "recording"
            ],
        )
        result[recorder_name] = {
            "aws.config.config_recorder.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



"""State module for managing EC2 VPC Peering Connection Options."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str,
    peer_allow_remote_vpc_dns_resolution: bool = None,
    peer_allow_classic_link_to_remote_vpc: bool = None,
    peer_allow_vpc_to_remote_classic_link: bool = None,
    allow_remote_vpc_dns_resolution: bool = None,
    allow_classic_link_to_remote_vpc: bool = None,
    allow_vpc_to_remote_classic_link: bool = None,
) -> Dict[str, Any]:
    """Requests to modify the VPC peering connection options.

    If the peered VPCs are in the same AWS account, you can enable DNS resolution for queries from the local VPC.
    This ensures that queries from the local VPC resolve to private IP addresses in the peer VPC.
    This option is not available if the peered VPCs are in different AWS accounts or different Regions.
    For peered VPCs in different AWS accounts,
    each AWS account owner must initiate a separate request to modify the peering connection options.
    For inter-region peering connections,
    you must use the Region for the requester VPC to modify the requester VPC peering options
    and the Region for the accepter VPC to modify the accepter VPC peering options.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, optional):
            An identifier of the resource in the provider. It is an identifier of the
            vpc peering connection of which the options are part of. Defaults to None.

        peer_allow_remote_vpc_dns_resolution(bool, optional):
            VPC peering connection option for the accepter:
            Requester Indicates whether a local VPC can resolve public DNS hostnames to private IP addresses
            when queried from instances in a peer VPC.

        peer_allow_classic_link_to_remote_vpc(bool, optional):
            VPC peering connection option for the accepter:
            Indicates whether a local ClassicLink connection can communicate with the peer VPC
            over the VPC peering connection.

        peer_allow_vpc_to_remote_classic_link(bool, optional):
            VPC peering connection option for the accepter:
            Indicates whether a local VPC can communicate with a ClassicLink connection in the peer VPC
            over the VPC peering connection.

        allow_remote_vpc_dns_resolution(bool, optional):
            VPC peering connection option for the requester:
            Requester Indicates whether a local VPC can resolve public DNS hostnames to private IP addresses
            when queried from instances in a peer VPC.

        allow_classic_link_to_remote_vpc(bool, optional):
            VPC peering connection option for the requester:
            Indicates whether a local ClassicLink connection can communicate with the peer VPC
            over the VPC peering connection.

        allow_vpc_to_remote_classic_link(bool, optional):
            VPC peering connection option for the requester:
            Indicates whether a local VPC can communicate with a ClassicLink connection in the peer VPC
            over the VPC peering connection.

    Request Syntax:
        .. code-block:: sls

            [vpc-peering-connection-id]:
              aws.ec2.vpc_peering_connection_options.present:
                - resource_id: "string"
                - peer_allow_remote_vpc_dns_resolution: True|False
                - peer_allow_classic_link_to_remote_vpc: True|False
                - peer_allow_vpc_to_remote_classic_link: True|False
                - allow_remote_vpc_dns_resolution: True|False
                - allow_classic_link_to_remote_vpc: True|False
                - allow_vpc_to_remote_classic_link: True|False

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            pcx-ae89ce9b:
                aws.ec2.vpc_peering_connection_options.present:
                    - resource_id: pcx-ae89ce9b
                    - peer_allow_classic_link_to_remote_vpc: true
                    - allow_vpc_to_remote_classic_link: true
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    if not resource_id:
        result["result"] = False
        result["comment"] = ("A valid resource id was not provided",)
        return result
    else:
        before = await hub.exec.aws.ec2.vpc_peering_connection_options.get(
            ctx, name=name, resource_id=resource_id
        )

        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = before["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        new_options = {
            "peer_allow_classic_link_to_remote_vpc": peer_allow_classic_link_to_remote_vpc,
            "peer_allow_remote_vpc_dns_resolution": peer_allow_remote_vpc_dns_resolution,
            "peer_allow_vpc_to_remote_classic_link": peer_allow_vpc_to_remote_classic_link,
            "allow_classic_link_to_remote_vpc": allow_classic_link_to_remote_vpc,
            "allow_remote_vpc_dns_resolution": allow_remote_vpc_dns_resolution,
            "allow_vpc_to_remote_classic_link": allow_vpc_to_remote_classic_link,
        }
        modify_options_ret = await hub.tool.aws.ec2.vpc_peering_connection_utils.update_vpc_peering_connection_options(
            ctx,
            resource_id,
            name,
            before,
            new_options,
        )

        result["result"] = modify_options_ret["result"]
        result["comment"] += modify_options_ret["comment"]
        if not modify_options_ret["result"]:
            return result

        options_updated = bool(modify_options_ret["ret"])
        resource_updated = resource_updated or options_updated
        if options_updated and ctx.get("test", False):
            for key, value in modify_options_ret["ret"].items():
                plan_state[key] = value

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    "aws.ec2.vpc_peering_connection_options", name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    "aws.ec2.vpc_peering_connection_options", name
                )
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                "aws.ec2.vpc_peering_connection_options", name
            )

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.ec2.vpc_peering_connection_options.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result

        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """A No-Op function for vpc_peering_connection_options.

    This is a configuration resource of the vpc_peering_connection resource.
    It's not possible to delete vpc_peering_connection_options,
    You need to delete the vpc_peering_connection resource by calling vpc_peering_connection.absent,
    while providing the vpc_peering_connection id.
    If you want to modify the vpc_peering_connection_options resource,
    use the vpc_peering_connection_options.present.

    Args:
        name:
            An Idem name of the resource.

        resource_id:
            The id of the vpc peering connection.

    Request Syntax:
        .. code-block:: sls

            [vpc-peering-connection-id]:
              aws.ec2.vpc_peering_connection_options.absent:
                - name: "string"

    Returns:
        Dict[str, Any]
    """
    result = dict(
        comment=(
            "No-Op: The VPC peering connection options can not be deleted as it is not an AWS resource",
        ),
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Describes one or more of your VPC peering connection options.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.vpc_peering_connection_options.py
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_vpc_peering_connections(ctx)

    if not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.ec2.vpc_peering_connection {ret['comment']}"
        )
        return {}

    for vpc_peering_connection in ret["ret"]["VpcPeeringConnections"]:
        resource_id = vpc_peering_connection.get("VpcPeeringConnectionId")
        vpc_peering_connection_options_translated = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_peering_connection_options_to_present(
            raw_resource=vpc_peering_connection, idem_resource_name=resource_id
        )

        result[vpc_peering_connection_options_translated["resource_id"]] = {
            "aws.ec2.vpc_peering_connection_options.present": [
                vpc_peering_connection_options_translated
            ]
        }

    return result



"""
Functions for interacting with an instance based on its ID
"""
import copy
import itertools
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


DEFAULT_SSH_PORT = 22
DEFAULT_TUNNEL_PLUGIN = "asyncssh"


def __init__(hub):
    hub.pop.sub.add(dyne_name="heist")


async def bootstrap(
    hub,
    ctx,
    instance_id,
    *,
    username: str = None,
    host: str = None,
    port: int = DEFAULT_SSH_PORT,
    ssh_public_key: str = None,
    ssh_private_key: str = None,
    availability_zone: str = None,
    heist_manager: str,
    artifact_version: str = None,
    tunnel_plugin: str = DEFAULT_TUNNEL_PLUGIN,
    auto_configure: bool = False,
    verify: bool = True,
    **kwargs,
):
    """Connect to the instance with an ssh keypair and call heist to bootstrap it.

    If an ssh keypair is not provided one will be created.

    Args:
        instance_id:
            An AWS EC2 Instance ID.

        username(str, Optional):
            The instance OS username to use in the connection, Defaults to ec2-user.

        host(str, Optional):
            The public ip address or dns name of the instance.  Defaults to autodetect.

        port(int, Optional):
            The port to connect to on the host.  Defaults to 22.

        ssh_public_key(str, Optional):
            A public ssh key or path to send to the instance.

        ssh_private_key(str, Optional):
            A private ssh key or path to send to the instance.

        availability_zone(str, Optional):
            The Availability Zone in which the EC2 instance was launched.

        heist_manager(str, Required):
            The heist manager to use to bootstrap the instance (I.E. salt.minion).

        artifact_version(str, Optional):
            The version of the heist_manager's artifact to upload to the instance, Defaults to latest.

        tunnel_plugin(str, Optional):
            The heist tunnel plugin to use to ssh into the instance. Defaults to asyncssh.

        auto_configure(Bool, Optional):
            automatically apply internet_gateway/subnet/vpc/route_table changes needed to SSH into an instance.

        verify(bool, Optional):
            If set to True, then verify that the instance is prepared for ssh connections.

        kwargs:
            All extra kwargs are used in asyncssh.SSHClientConnectionOptions.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    Examples:
        .. code-block:: bash

            idem exec aws.ec2.instance.bootstrap <instance_id> heist_manager="salt.minion"
    """
    result = dict(comment=[], result=True, ret=None)

    if verify or auto_configure:
        verify_ssh_ret = await hub.exec.aws.ec2.instance.verify_ssh(
            ctx, instance_id=instance_id
        )
        result["result"] &= verify_ssh_ret.result
        result["comment"] += verify_ssh_ret.comment
        if not result["result"]:
            return result

    with hub.tool.aws.ec2.instance.key_pair.verify(
        ssh_public_key=ssh_public_key,
        ssh_private_key=ssh_private_key,
        auto_configure=auto_configure,
    ) as key_pair:
        verified_public_key, private_key_file = key_pair

        connection_ret = await hub.exec.aws.ec2.instance.connect(
            ctx,
            instance_id=instance_id,
            username=username,
            host=host,
            port=port,
            tunnel_plugin=tunnel_plugin,
            ssh_public_key=verified_public_key,
            private_key_file=private_key_file,
            availability_zone=availability_zone,
            **kwargs,
        )
        result["result"] &= connection_ret.result
        result["comment"] += connection_ret.comment
        if not result["result"]:
            return result

        # Bootstrap the instance
        await hub.heist[heist_manager].run(
            remotes={instance_id: connection_ret.ret}, artifact_version=artifact_version
        )
        # TODO Check for a successful bootstrapping

    result["comment"] += [
        f"Successfully bootstrapped instance '{instance_id}' with '{heist_manager}'",
    ]

    return result


async def verify_ssh(hub, ctx, instance_id: str, auto_configure: bool = False):
    """Verify that an instance meets all the requirements for SSH connection

    Args:
        instance_id(str):
            An AWS EC2 Instance ID.

        auto_configure(bool, Optional):
            Automatically apply internet_gateway/subnet/vpc/route_table changes needed to SSH into an instance.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    Examples:

        .. code-block:: bash

            idem exec aws.ec2.instance.verify_ssh <instance_id>
    """
    result = dict(result=True, comment=[], ret=None)

    # Create an instance resource object to use for defaults
    instance = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", instance_id)

    # Verify that the instance is in a state with all the right permissions
    if not instance.state["Name"] == "running":
        result["comment"] += ["Instance is not running"]
        if not auto_configure:
            result["result"] = False
            return result
        # TODO start the instance

    if not instance.vpc.vpc_id:
        result["comment"] += [f"No vpc attached to instance"]
        if not auto_configure:
            result["result"] = False
            return result
        # TODO attach the default vpc

    if not instance.subnet.subnet_id:
        result["comment"] += [f"No subnet attached to instance"]
        if not auto_configure:
            result["result"] = False
            return result
        # TODO attach the default subnet for the instance's availability zone

    # Verify that an internet gateway is attached to the instance's vpc
    internet_gateway_ret = await hub.exec.boto3.client.ec2.describe_internet_gateways(
        ctx,
        Filters=[
            {"Name": "attachment.vpc-id", "Values": [instance.vpc.vpc_id]},
        ],
    )
    if not internet_gateway_ret.result and internet_gateway_ret.ret["InternetGateways"]:
        result["comment"] += [
            f"No internet gateway attached to instance's vpc: {instance.vpc.vpc_id}"
        ]
        if not auto_configure:
            result["result"] = False
        # TODO attach an internet gateway to the vpc

    internet_gateway_ids = [
        igw["InternetGatewayId"] for igw in internet_gateway_ret.ret["InternetGateways"]
    ]

    # Verify that there is a route with 0.0.0.0/0 as the destination and the internet gateway for your VPC as the target
    ipv4_route_table_ret = await hub.exec.boto3.client.ec2.describe_route_tables(
        ctx,
        Filters=[
            {"Name": "vpc-id", "Values": [instance.vpc.vpc_id]},
            {"Name": "route.destination-cidr-block", "Values": ["0.0.0.0/0"]},
            {"Name": "route.gateway-id", "Values": internet_gateway_ids},
        ],
    )
    has_route_table = False
    if not (ipv4_route_table_ret and ipv4_route_table_ret.ret["RouteTables"]):
        result["comment"] += [
            f"No route from vpc's internet gateway {internet_gateway_ids} to '0.0.0.0/0'"
        ]
    else:
        has_route_table = True

    if not has_route_table:
        # Verify that there is a route with ::/0 as the destination and the internet gateway for your VPC as the target
        ipv6_route_table_ret = await hub.exec.boto3.client.ec2.describe_route_tables(
            ctx,
            Filters=[
                {"Name": "vpc-id", "Values": [instance.vpc.vpc_id]},
                {"Name": "route.destination-ipv6-cidr-block", "Values": ["::/0"]},
                {"Name": "route.gateway-id", "Values": internet_gateway_ids},
            ],
        )
        if not (ipv6_route_table_ret and ipv6_route_table_ret.ret["RouteTables"]):
            result["comment"] += [
                f"No route from vpc's internet gateway {internet_gateway_ids} to '::/0'"
            ]
        else:
            has_route_table = True

    # If there is still no route table then add a command to the comments that would enable it
    if not has_route_table:
        if not auto_configure:
            # At least one ipv4 or ipv6 route table must be configured for SSH
            result["result"] = False
        # TODO configure route table

        # Get the first internet gateway to use in an example
        igw_id = next(iter(internet_gateway_ids))
        # Get the id of the main route table attached to the vpc
        try_route_table = await hub.exec.boto3.client.ec2.describe_route_tables(
            ctx,
            Filters=[
                {"Name": "vpc-id", "Values": [instance.vpc.vpc_id]},
                {"Name": "route.gateway-id", "Values": [igw_id]},
            ],
        )
        if try_route_table and try_route_table.ret:
            result["comment"] += [
                "Use the following command to create a valid SSH route for the instance"
            ]
            # Get the id of the main route table
            rt_id = next(iter(try_route_table.ret["RouteTables"]))["RouteTableId"]
            result["comment"] += [
                f"idem exec boto3.client.ec2.create_route RouteTableID={rt_id} VpcEndpointId={instance.vpc.vpc_id} DestinationCidrBlock='0.0.0.0/0' GatewayId={igw_id}"
            ]

    # Verify that the ACL for the subnet allows inbound traffic
    network_acl_ret = await hub.exec.boto3.client.ec2.describe_network_acls(
        ctx,
        Filters=[
            {"Name": "association.subnet-id", "Values": [instance.subnet.subnet_id]},
            {"Name": "entry.rule-action", "Values": ["allow"]},
            {"Name": "vpc-id", "Values": [instance.vpc.vpc_id]},
        ],
    )
    if not (network_acl_ret and network_acl_ret.ret["NetworkAcls"]):
        result["comment"] += [f"No valid ACL"]
        if not auto_configure:
            result["result"] = False
        # TODO configure network acl

    return result


async def connect(
    hub,
    ctx,
    instance_id: str,
    *,
    username: str = None,
    host: str = None,
    port: int = DEFAULT_SSH_PORT,
    tunnel_plugin: str = DEFAULT_TUNNEL_PLUGIN,
    ssh_public_key: str,
    private_key_file: str,
    availability_zone: str = None,
    **kwargs,
):
    """Attempt connection with SSH to the instance with the given parameters.

    If no host or username is specified, then make reasonable guesses and return the target configuration that was successful

    Args:
        instance_id(str):
            An AWS EC2 Instance ID.

        username(str, Optional):
            The instance OS username to use in the connection, Defaults to ec2-user.

        host(str, Optional):
            The public ip address or dns name of the instance.  Defaults to autodetect.

        port(int, Optional):
            The port to connect to on the host.  Defaults to 22.

        ssh_public_key(str, Optional):
            A public ssh key or path to send to the instance.

        private_key_file(str, Optional):
            The path to a private ssh key file.

        availability_zone(str, Optional):
            The Availability Zone in which the EC2 instance was launched.

        tunnel_plugin(str, Optional):
            The heist tunnel plugin to use to ssh into the instance. Defaults to asyncssh.

        kwargs:
            All extra kwargs are used in asyncssh.SSHClientConnectionOptions.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    Examples:
        .. code-block:: bash

            idem exec aws.ec2.instance.connect <instance_id>
    """
    result = dict(result=True, comment=[], ret={})

    # Create an instance resource object to use for defaults
    instance = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", instance_id)

    # Use the availability zone defined in the instance if none was provided
    if availability_zone is None:
        availability_zone = instance.placement["AvailabilityZone"]

    # If no host was provided then try all the instance's dns names and ip addresses
    if host:
        hosts_to_try = (host,)
    else:
        hosts_to_try = (
            instance.public_dns_name,
            instance.private_dns_name,
            instance.private_ip_address,
            instance.ipv6_address,
            instance.public_ip_address,
        )

    # If no username was provided then try all common known usernames for ec2 images
    if username:
        # Only use the given username
        user_names_to_try = (username,)
    else:
        user_names_to_try = (
            "ec2-user",
            "root",
            "admin",
            "ubuntu",
            "bitnami",
            "fedora",
            "centos",
        )

    # If no username or host was specified, then try multiple combinations until one works
    for try_user, try_host in itertools.product(user_names_to_try, hosts_to_try):
        if not try_host:
            # the ipv6_address or other attempted host name might not be set
            continue
        target = dict(
            host=try_host,
            port=port,
            bootstrap=True,
            username=try_user,
            client_keys=[private_key_file],
            IdentitiesOnly="yes",
            **kwargs,
        )

        # Connect to the ec2 instance with the public key, connection will be available for 60 seconds
        connection_ret = await hub.exec.boto3.client[
            "ec2-instance-connect"
        ].send_ssh_public_key(
            ctx,
            InstanceId=instance_id,
            InstanceOSUser=try_user,
            SSHPublicKey=ssh_public_key,
            AvailabilityZone=availability_zone,
        )
        if connection_ret.comment:
            hub.log.debug(
                f"Sending public key to {instance_id}: {connection_ret.comment}"
            )

        if not connection_ret.result:
            hub.log.debug(f"Connection error sending public key to {instance_id}")
            continue
        if not connection_ret.ret.get("Success"):
            hub.log.debug(f"Unable to send public key to {instance_id}")
            continue

        hub.log.debug(
            f"Connecting to instance '{instance_id}' with ssh: {try_user}@{try_host}"
        )

        # Create a connection verify the ability to SSH to the instance
        ret = await hub.tunnel[tunnel_plugin].create(
            name=instance_id, target=copy.deepcopy(target)
        )
        if not ret:
            hub.log.debug(
                f"Unable to connect to the instance with SSH: {try_user}@{try_host}"
            )
            continue

        # Success!
        result["comment"] += [
            f"Successfully connected to instance with SSH: {try_user}@{try_host}"
        ]
        result["ret"] = target
        return result

    # There were no successful connections in the loop
    result["comment"] += [
        f"Unable to connect to instance with SSH",
        "Try troubleshooting: https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/TroubleshootingInstancesConnecting.html",
    ]
    result["result"] = False

    return result


async def start(hub, ctx, instance_id: str):
    """
    Start the instance and wait for it to be running
    """
    result = dict(result=True, comment=[], ret={})

    ret = await hub.exec.boto3.client.ec2.start_instances(
        ctx, InstanceIds=[instance_id]
    )
    if ret.comment:
        result["comments"].append(ret.comment)
    if not ret.result:
        return False

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", instance_id)
    await hub.tool.boto3.resource.exec(ctx, resource, "wait_until_running")

    return result


async def stop(hub, ctx, instance_id: str):
    """
    Stop the instance and wait for it to be stopped.
    """
    result = dict(result=True, comment=[], ret={})

    ret = await hub.exec.boto3.client.ec2.stop_instances(
        ctx, InstanceIds=[instance_id], Hibernate=False, Force=False
    )
    if ret.comment:
        result["comments"].append(ret.comment)
    if not ret.result:
        return False

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", instance_id)
    await hub.tool.boto3.resource.exec(ctx, resource, "wait_until_stopped")

    return result


async def get(
    hub,
    ctx,
    *,
    name=None,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """
    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, Optional):
            AWS VPC id to identify the resource.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_instances.
    """
    result = dict(comment=[], ret=None, result=True)
    if resource_id:
        ret = await hub.exec.boto3.client.ec2.describe_instances(
            ctx, InstanceIds=[resource_id], Filters=filters
        )
    else:
        ret = await hub.exec.boto3.client.ec2.describe_instances(
            ctx,
            Filters=filters,
        )

    # If there was an error in the call then report failure
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    present_states = await hub.tool.aws.ec2.instance.state.convert_to_present(
        ctx, ret.ret
    )

    # If the resource can't be found but there were no results then "result" is True and "ret" is None
    if not present_states:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.instance", name=name
            )
        )
        return result

    # return the first result as a plain dictionary
    result["ret"] = next(iter((present_states).values()))

    return result


async def list_(hub, ctx, *, name=None, filters: List = None) -> Dict:
    """
    Args:
        name(str):
            The name of the Idem state.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_instances.
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.ec2.describe_instances(
        ctx,
        Filters=filters,
    )

    # If there was an error in the call then report failure
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    present_states = await hub.tool.aws.ec2.instance.state.convert_to_present(
        ctx, ret.ret
    )
    if not present_states:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.instance", name=name
            )
        )
        return result

    # Return a list of dictionaries with details about all the instances
    result["ret"] = list(present_states.values())

    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: List[Dict[str, Any]] or Dict[str, Any],
    new_tags: List[Dict[str, Any]] or Dict[str, Any],
):
    """Update tags of WAFV2 resources

    Args:
        resource_arn(str): Amazon Resource Identifier(ARN) of WAFV2 resource
        old_tags(dict): list of old tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
         {tag-key: tag-value}
        new_tags(dict): list of new tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
         {tag-key: tag-value}. If this value is None, the function will do no operation on tags.


    Returns:
        {"result": True|False, "comment": "A message", "ret": dict of updated tags}

    """

    tags_to_add = {}
    tags_to_remove = {}
    if isinstance(old_tags, List):
        old_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(old_tags)
    if isinstance(new_tags, List):
        new_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(new_tags)
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result

    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.wafv2.untag_resource(
                ctx, ResourceARN=resource_arn, TagKeys=list(tags_to_remove)
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.wafv2.tag_resource(
                ctx,
                ResourceARN=resource_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = new_tags
    result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
        tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
    )
    return result



from typing import Dict


async def get(hub, ctx, resource_id, name) -> Dict:
    """Returns the ACL policy for the S3 bucket.

    Args:
        resource_id(str): AWS S3 bucket name.
        name(str): The name of the Idem state.
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.s3.get_bucket_acl(ctx, Bucket=resource_id)
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.s3.bucket_acl", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result["ret"] = hub.tool.aws.s3.conversion_utils.convert_raw_acl_policy_to_present(
        raw_resource=ret["ret"], bucket_name=resource_id
    )
    return result



"""State module for managing Amazon Organization Units."""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.organizations.account.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.organization.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    parent_id: str,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Creates an organizational unit (OU) within a root or parent OU.

    An OU is a container for accounts that enables you to organize your accounts to apply policies according
    to your business requirements. The number of levels deep that you can nest OUs is dependent upon the policy
    types enabled for that root. For service control policies, the limit is five.

    Args:
        name(str):
            An Idem name of the resource.
        parent_id(str):
            The unique identifier (ID) of the parent root or OU that you want to create the new OU in.
        resource_id(str, Optional):
            The ID of the organization unit in Amazon Web Services.
        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the organization unit.

            * Key (*str*):
                The key identifier, or name, of the tag.
            * Value (*str*):
                The string value that's associated with the key of the tag.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_organization_unit]:
          aws.organizations.organization_unit.present:
          - name: 'string'
          - resource_id: 'string'
          - parent_id: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_organization_unit:
              aws.organizations.organization_unit.present:
                - name: 'idem_test_organization_unit'
                - parent_id: 'o-parent-id'
                - tags:
                  - Key: 'provider'
                    Value: 'idem'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    existing_tags = None

    if resource_id:
        before = await hub.exec.boto3.client.organizations.describe_organizational_unit(
            ctx, OrganizationalUnitId=resource_id
        )
    update = False
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if not before:
        # organization unit does not exist, create

        try:

            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "parent_id": parent_id,
                        "name": name,
                        "tags": tags,
                    },
                )
                result["comment"] = result["comment"] + (
                    f"Would create aws.organizations.organization_unit '{name}'.",
                )
                return result

            create_ret = (
                await hub.exec.boto3.client.organizations.create_organizational_unit(
                    ctx,
                    ParentId=parent_id,
                    Name=name,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
            )

            result["result"] = create_ret["result"]

            if not result["result"]:
                result["comment"] = result["comment"] + create_ret["comment"]
                return result
            resource_id = create_ret["ret"]["OrganizationalUnit"]["Id"]
            result["comment"] = result["comment"] + (
                f"Created aws.organizations.organization_unit '{name}'.",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        # organizational_unit exists , update
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would update aws.organizations.organization_unit '{name}'.",
            )
        existing_ou_name = before["ret"]["OrganizationalUnit"]["Name"]

        old_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=resource_id
        )
        if not old_tags["result"]:
            hub.log.debug(
                f"Unable to list tags for resource {resource_id} with error: {old_tags['comment']}"
            )
            result["comment"] = result["comment"] + old_tags["comment"]
            result["result"] = old_tags["result"]
        else:
            existing_tags = old_tags["ret"]["Tags"]

        result[
            "old_state"
        ] = hub.tool.aws.organizations.conversion_utils.convert_raw_ou_to_present(
            before["ret"]["OrganizationalUnit"], parent_id, existing_tags
        )
        plan_state = copy.deepcopy(result["old_state"])

        try:

            if existing_ou_name != name:

                if ctx.get("test", False):
                    plan_state["name"] = name
                else:
                    update_ret = await hub.exec.boto3.client.organizations.update_organizational_unit(
                        ctx,
                        OrganizationalUnitId=resource_id,
                        Name=name,
                    )
                    result["result"] = update_ret["result"]
                    if not result["result"]:
                        result["comment"] = result["comment"] + update_ret["comment"]
                        return result
                    result["comment"] = result["comment"] + (
                        f"Updated ou name on aws.organizations.organization_unit '{name}'.",
                    )

                update = True
            if tags is not None and tags != result["old_state"].get("tags"):
                update_tags_ret = await hub.tool.aws.organizations.tag.update_tags(
                    ctx, resource_id, result["old_state"].get("tags"), tags
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = False
                    return result

                elif update_tags_ret["ret"] is not None:
                    update = True
                    result["comment"] = result["comment"] + (
                        f"Updated tags on aws.organizations.organization_unit '{name}'.",
                    )

                if ctx.get("test", False) and update_tags_ret["ret"]:
                    plan_state["tags"] = update_tags_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update aws.organizations.organization_unit '{name}'",
                    )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:

        if ctx.get("test", False):
            result["new_state"] = plan_state
            return result

        if not before or update:

            after = (
                await hub.exec.boto3.client.organizations.describe_organizational_unit(
                    ctx, OrganizationalUnitId=resource_id
                )
            )

            if after.get("ret"):
                updated_tag = (
                    await hub.exec.boto3.client.organizations.list_tags_for_resource(
                        ctx, ResourceId=resource_id
                    )
                )
                if not updated_tag["result"]:
                    hub.log.debug(
                        f"Unable to list tags for resource {resource_id} with error: {updated_tag['comment']}"
                    )
                    result["comment"] = result["comment"] + updated_tag["comment"]
                    result["result"] = updated_tag["result"]
                result[
                    "new_state"
                ] = hub.tool.aws.organizations.conversion_utils.convert_raw_ou_to_present(
                    after["ret"]["OrganizationalUnit"],
                    parent_id,
                    updated_tag["ret"].get("Tags")
                    if updated_tag and updated_tag.get("ret")
                    else None,
                )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    """Deletes an organizational unit (OU) from a root or another OU.

    You must first remove all accounts and child OUs from the OU that you want to delete.
    This operation can be called only from the organization's management account.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            The ID of the organization unit in Amazon Web Services.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_organization_unit]:
          aws.organizations.organization_unit.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_organization_unit:
              aws.organizations.organization_unit.absent:
                - name: 'idem_test_organization_unit'
                - resource_id: 'ou-rootid-ouid'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_organizational_unit(
        ctx, OrganizationalUnitId=resource_id
    )

    if not before:
        result["comment"] = result["comment"] + (
            f"aws.organizations.organization_unit '{name}' already absent",
        )
    else:

        result[
            "old_state"
        ] = hub.tool.aws.organizations.conversion_utils.convert_raw_ou_to_present(
            before["ret"]["OrganizationalUnit"]
        )

        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.organizations.organization_unit '{name}'",
            )
            return result
        try:
            ret = await hub.exec.boto3.client.organizations.delete_organizational_unit(
                ctx, OrganizationalUnitId=resource_id
            )

            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result["comment"] + (
                f"aws.organizations.organization_unit '{name}' deleted.",
            )
            result["result"] = ret["result"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS Organizations Units (OUs) in a way that can be recreated/managed with the corresponding "present" function.

    Idem does not support passing arguments to the describe function as of today, hence describe organization_unit
    will describe all the organizational units one level under the root of the organization.
    In future if Idem starts supporting input arguments, we can pass any parent_id and this function will list
    all organizational units one level under the parent.

    Returns:
        Dict[str, Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.organizations.organization_unit
    """
    result = {}

    list_roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

    if not (list_roots_resp and list_roots_resp["ret"].get("Roots")):
        return result

    parent_id = list_roots_resp["ret"]["Roots"][0]["Id"]

    if parent_id is None:
        return result

    org_units = (
        await hub.exec.boto3.client.organizations.list_organizational_units_for_parent(
            ctx, ParentId=parent_id
        )
    )

    if not org_units["result"]:
        hub.log.debug(
            f"Could not describe organization units for parent_id {parent_id} with error: {org_units['comment']}"
        )
        return {}

    organizational_units = org_units["ret"]["OrganizationalUnits"]

    for ou in organizational_units:
        tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=ou["Id"]
        )
        if not tags["result"]:
            hub.log.debug(
                f"Unable to list tags for resource {ou['Id']} with error: {tags['comment']}"
            )
            continue
        translated_resource = (
            hub.tool.aws.organizations.conversion_utils.convert_raw_ou_to_present(
                ou, parent_id, tags["ret"].get("Tags") if tags else None
            )
        )

        result[ou["Name"]] = {
            "aws.organizations.organization_unit.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



from typing import List


async def apply(
    hub, ctx, resource, *, old_value, new_value, comments: List[str]
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    if not old_value:
        # Associate the new iam instance profile
        ret = await hub.exec.boto3.client.ec2.associate_iam_instance_profile(
            ctx,
            InstanceId=resource.id,
            IamInstanceProfile={
                "Arn": new_value,
            },
        )
        if ret.comment:
            comments.append(ret.comment)
        return ret.result

    # Get the association id
    ret = await hub.exec.boto3.client.ec2.describe_iam_instance_profile_associations(
        ctx, Filters=[{"Name": "instance-id", "Values": [resource.id]}]
    )
    if ret.comment:
        comments.append(ret.comment)
    if not ret:
        return False

    association_id = None
    for association in ret.ret["IamInstanceProfileAssociations"]:
        if association["State"] == "associated":
            association_id = association["AssociationId"]
            break

    if not new_value:
        # disassociate the previous iam profile
        if association_id:
            ret = await hub.exec.boto3.client.ec2.associate_iam_instance_profile(
                ctx, AssociationId=association_id
            )
            if ret.comment:
                comments.append(ret.comment)
            return ret.result
        comments.append(f"Iam profile is already disassociated")
        return True
    else:
        # Replace the old iam instance profile with the new one
        ret = await hub.exec.boto3.client.ec2.replace_iam_instance_profile(
            ctx,
            InstanceId=resource.id,
            IamInstanceProfile=dict(Arn=new_value, AssociationId=association_id),
        )
        if ret.comment:
            comments.append(ret.comment)
        return ret.result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.sqs.create_queue
hub.exec.boto3.client.sqs.delete_queue
hub.exec.boto3.client.sqs.list_queues
hub.exec.boto3.client.sqs.purge_queue
hub.exec.boto3.client.sqs.tag_queue
hub.exec.boto3.client.sqs.untag_queue
resource = hub.tool.boto3.resource.create(ctx, "sqs", "Queue", resource_id)
hub.tool.boto3.resource.exec(resource, add_permission, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, change_message_visibility_batch, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete_messages, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, purge, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, receive_messages, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, remove_permission, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, send_message, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, send_messages, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, set_attributes, *args, **kwargs)
"""
import copy
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    delay_seconds: int = None,
    maximum_message_size: int = None,
    message_retention_period: int = None,
    policy: Dict[str, Any] = None,
    receive_message_wait_time_seconds: int = None,
    redrive_policy: make_dataclass(
        "RedrivePolicy", [("deadLetterTargetArn", str), ("maxReceiveCount", int)]
    ) = None,
    visibility_timeout: int = None,
    kms_master_key_id: str = None,
    kms_data_key_reuse_period_seconds: int = None,
    sqs_managed_sse_enabled: bool = None,
    fifo_queue: bool = None,
    content_based_deduplication: bool = None,
    deduplication_scope: str = None,
    fifo_throughput_limit: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new standard or FIFO queue. You can pass one or more attributes in the request. Keep the following in
    mind:
        If you don't specify the fifo_queue attribute, Amazon SQS creates a standard queue. You can't change the
        queue type after you create it, and you can't convert an existing standard queue into a FIFO queue. You must
        either create a new FIFO queue for your application or delete your existing standard queue and recreate it as a
        FIFO queue. For more information, see Moving From a Standard Queue to a FIFO Queue in the Amazon SQS Developer
        Guide. If you don't provide a value for an attribute, the queue is created with the default value for the
        attribute. If you delete a queue, you must wait at least 60 seconds before creating a queue with the same
        name. To successfully create a new queue, you must provide a queue name that adheres to the limits related to
        queues and is unique within the scope of your queues. After you create a queue, you must wait at least one
        second after the queue is created to be able to use the queue. Cross-account permissions
        don't apply to this action. For more information, see Grant cross-account permissions to a role and a user name
        in the Amazon SQS Developer Guide.

    Args:
        name(str): The name of the new queue. The following limits apply to this name:
            A queue name can have up to 80 characters. Valid values: alphanumeric characters, hyphens (-), and
            underscores (_). A FIFO queue name must end with the .fifo suffix. Case-sensitive.
        resource_id(str, optional): The URL of the Amazon SQS queue. Case-sensitive. Defaults to None.
        delay_seconds(int, optional): The length of time, in seconds, for which the delivery of all messages in the
            queue is delayed. Valid values: An integer from 0 to 900 seconds (15 minutes). Default: 0.
        maximum_message_size(int, optional): The limit of how many bytes a message can contain before
            Amazon SQS rejects it. Valid values: An integer from 1,024 bytes (1 KiB) to 262,144 bytes (256 KiB).
            Default: 262,144 (256 KiB).
        message_retention_period(int, optional): The length of time, in seconds, for which Amazon SQS retains a message.
            Valid values: An integer from 60 seconds (1 minute) to 1,209,600 seconds (14 days).
            Default: 345,600 (4 days).
        policy(Dict, optional): The queue's policy. A valid Amazon Web Services policy. For more information
            about policy structure, see Overview of Amazon Web Services IAM Policies in the Amazon IAM User Guide.
        receive_message_wait_time_seconds(int, optional): The length of time, in seconds, for which a ReceiveMessage
            action waits for a message to arrive. Valid values: An integer from 0 to 20 (seconds). Default: 0.
        redrive_policy(Dict, optional): The dictionary that includes the parameters for the dead-letter queue
            functionality of the source queue. For more information about the redrive policy and dead-letter queues,
            see Using Amazon SQS Dead-Letter Queues in the Amazon SQS Developer Guide.
            * deadLetterTargetArn (str): The Amazon Resource Name (ARN) of the dead-letter queue to which Amazon SQS
                moves messages after the value of maxReceiveCount is exceeded.
            * maxReceiveCount (int): The number of times a message is delivered to the source queue before being moved to
                the dead-letter queue. When the ReceiveCount for a message exceeds the maxReceiveCount for a queue,
                Amazon SQS moves the message to the dead-letter-queue. The dead-letter queue of a FIFO queue must also
                be a FIFO queue. Similarly, the dead-letter queue of a standard queue must also be a standard queue.
        visibility_timeout(int, optional): The visibility timeout for the queue, in seconds. Valid values: An integer
            from 0 to 43,200 (12 hours). Default: 30. For more information about the visibility timeout,
            see Visibility Timeout in the Amazon SQS Developer Guide.
        The following attributes apply only to server-side-encryption:
        kms_master_key_id(str, optional): The ID of an Amazon Web Services managed customer master key (CMK) for
            Amazon SQS or a custom CMK. For more information, see Key Terms. While the alias of the Amazon Web Services
            managed CMK for Amazon SQS is always alias/aws/sqs, the alias of a custom CMK can, for example, be
            alias/MyAlias . For more examples, see KeyId in the Key Management Service API Reference.
        kms_data_key_reuse_period_seconds(int, optional): The length of time, in seconds, for which Amazon SQS can
            reuse a data key to encrypt or decrypt messages before calling KMS again. An integer representing seconds,
            between 60 seconds (1 minute) and 86,400 seconds (24 hours). Default: 300 (5 minutes). A shorter time period
            provides better security but results in more calls to KMS which might incur charges after Free Tier.
            For more information, see How Does the Data Key Reuse Period Work?
        sqs_managed_sse_enabled(bool, optional): enables server-side queue encryption using SQS owned encryption keys.
            Only one server-side encryption option is supported per queue.

        The following attributes apply only to FIFO (first-in-first-out) queues:
        fifo_queue(bool, optional): Designates a queue as FIFO. Valid values are true and false. If you don't specify
            the FifoQueue attribute, Amazon SQS creates a standard queue. You can provide this attribute only during
            queue creation. You can't change it for an existing queue. When you set this attribute, you must also
            provide the MessageGroupId for your messages explicitly. For more information, see FIFO queue logic in
            the Amazon SQS Developer Guide.
        content_based_deduplication(bool, optional): Enables content-based deduplication. Valid values are true and
            false. For more information, see Exactly-once processing in the Amazon SQS Developer Guide. Note
            the following:
                Every message must have a unique MessageDeduplicationId.
                    You may provide a MessageDeduplicationId explicitly.
                    If you aren't able to provide a MessageDeduplicationId and you enable ContentBasedDeduplication
                        for your queue, Amazon SQS uses a SHA-256 hash to generate the MessageDeduplicationId using
                        the body of the message (but not the attributes of themessage).
                    If you don't provide a MessageDeduplicationId and the queue doesn't have
                        ContentBasedDeduplication set, the action fails with an error.
                    If the queue has ContentBasedDeduplication set, your MessageDeduplicationId overrides
                        the generated one.
                When ContentBasedDeduplication is in effect, messages with identical content sent within
                    the deduplication interval are treated as duplicates
                    and only one copy of the message is delivered.
                If you send one message with ContentBasedDeduplication enabled and then another message with a
                    MessageDeduplicationId that is the same as the one generated for the first
                    MessageDeduplicationId, the two messages are treated as duplicates and only one copy of the
                    message is delivered.

        The following attributes apply only to high throughput for FIFO queues:
        deduplication_scope(str, optional): Specifies whether message deduplication occurs at the message group
            or queue level. Valid values are messageGroup and queue.
        fifo_throughput_limit(str, optional): Specifies whether the FIFO queue throughput quota applies to the entire
            queue or per message group. Valid values are perQueue and perMessageGroupId. The perMessageGroupId value is
            allowed only when the value for DeduplicationScope is messageGroup.
            To enable high throughput for FIFO queues, do the following:
                Set DeduplicationScope to messageGroup.
                Set FifoThroughputLimit to perMessageGroupId.
            If you set these attributes to anything other than the values shown for
                enabling high throughput, normal throughput is in effect and deduplication occurs as specified.
            For information on throughput quotas, see Quotas related to messages in the Amazon SQS Developer
                Guide. Defaults to None.
        tags(Dict, optional): Add cost allocation tags to the specified Amazon SQS queue. For an overview, see Tagging
            Your Amazon SQS Queues in the Amazon SQS Developer Guide. When you use queue tags, keep the following
            guidelines in mind:
                Adding more than 50 tags to a queue isn't recommended.
                Tags don't have any semantic meaning. Amazon SQS interprets tags as character strings.
                Tags are case-sensitive.
                A new tag with a key identical to that of an existing tag overwrites the existing tag.
            For a full list of tag restrictions, see Quotas related to queues in the Amazon SQS Developer Guide.
            To be able to tag a queue on creation, you must have the sqs:CreateQueue and sqs:TagQueue permissions.
            Cross-account permissions don't apply to this action. For more information, see Grant cross-account
            permissions to a role and a user name in the Amazon SQS Developer Guide. Defaults to None.

    Request Syntax:
        [queue_name]:
          aws.sqs.queue.present:
            - delay_seconds: 'int'
            - maximum_message_size: 'int'
            - message_retention_period: 'int'
            - policy: 'dict'
            - receive_message_wait_time_seconds: 'int'
            - redrive_policy: 'dict'
            - visibility_timeout: 'int'
            - kms_master_key_id: 'string'
            - kms_data_key_reuse_period_seconds: 'int'
            - sqs_managed_sse_enabled: 'bool'
            - fifo_queue: 'bool'
            - content_based_deduplication: 'bool'
            - deduplication_scope: 'string'
            - fifo_throughput_limit: 'string'
            - tags:
                'string': 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            example_fifo_queue.fifo:
              aws.sqs.queue.present:
                - delay_seconds: 35
                - maximum_message_size: 12345
                - message_retention_period: 1759
                - policy:
                    Version: 2022-03-30
                    Id: Example_Queue_Policy
                    Statement:
                      Effect: Deny
                      Principal:
                        AWS:
                         - "000000000000"
                      Action: sqs:SendMessage
                      Resource: arn:aws:sqs:us-west-1:000000000000:example_sqs_queue.fifo
                - receive_message_wait_time_seconds: 3
                - redrive_policy:
                    deadLetterTargetArn: arn:aws:sqs:us-west-1:000000000000:dead_letter_sqs_queue.fifo
                    maxReceiveCount: 5
                - visibility_timeout: 75
                - kms_master_key_id: alias/sample/key
                - kms_data_key_reuse_period_seconds: 735
                - fifo_queue: true
                - content_based_deduplication: true
                - deduplication_scope: message_group
                - fifo_throughput_limit: perMessageGroupId
                - tags:
                    something: this-is-just-some-text
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    before = None
    queue_url = resource_id

    # Try to get the state of the resource if such is available
    if queue_url:
        queue_state_ret = await hub.exec.aws.sqs.queue.get(ctx, queue_url)
        if not queue_state_ret["result"] or not queue_state_ret["ret"]:
            result["result"] = False
            result["comment"] += queue_state_ret["comment"]
            return result

        before = queue_state_ret["ret"]

    present_attributes = {
        "delay_seconds": delay_seconds,
        "maximum_message_size": maximum_message_size,
        "message_retention_period": message_retention_period,
        "policy": policy,
        "receive_message_wait_time_seconds": receive_message_wait_time_seconds,
        "redrive_policy": redrive_policy,
        "visibility_timeout": visibility_timeout,
        "kms_master_key_id": kms_master_key_id,
        "kms_data_key_reuse_period_seconds": kms_data_key_reuse_period_seconds,
        "sqs_managed_sse_enabled": sqs_managed_sse_enabled,
        "fifo_queue": fifo_queue,
        "content_based_deduplication": content_based_deduplication,
        "deduplication_scope": deduplication_scope,
        "fifo_throughput_limit": fifo_throughput_limit,
    }

    # Convert the new state attributes to a dict of string values with raw attribute names
    plan_state_raw_attributes = (
        hub.tool.aws.sqs.conversion_utils.convert_present_attributes_to_raw(
            **present_attributes
        )
    )

    if before:

        # Check if the resource needs to be updated

        result["old_state"] = before

        # Convert the old state attributes to a dict of string values with raw attribute names
        old_state_raw_attributes = (
            hub.tool.aws.sqs.conversion_utils.convert_present_attributes_to_raw(
                before.get("delay_seconds"),
                before.get("maximum_message_size"),
                before.get("message_retention_period"),
                before.get("policy"),
                before.get("receive_message_wait_time_seconds"),
                before.get("redrive_policy"),
                before.get("visibility_timeout"),
                before.get("kms_master_key_id"),
                before.get("kms_data_key_reuse_period_seconds"),
                before.get("sqs_managed_sse_enabled"),
                before.get("fifo_queue"),
                before.get("content_based_deduplication"),
                before.get("deduplication_scope"),
                before.get("fifo_throughput_limit"),
            )
        )

        # Get only the attributes that need to be updated
        attributes_to_update = dict(
            plan_state_raw_attributes.items() - old_state_raw_attributes.items()
        )

        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            before.get("tags"), tags
        )

        resource_updated = attributes_to_update or tags_to_add or tags_to_remove

        # Handle the case when updating the resource with --test
        if ctx.get("test", False):

            # Construct the new state in the case of updating the resource with --test
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    **present_attributes,
                    "tags": tags,
                },
            )
            if attributes_to_update:
                # Updated resource
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    "aws.sqs.queue", name
                )

            if tags_to_add or tags_to_remove:
                # Updated tags
                result[
                    "comment"
                ] += hub.tool.aws.comment_utils.would_update_tags_comment(
                    tags_to_remove, tags_to_add
                )
        else:
            # Update the attributes
            if attributes_to_update:
                updated_attributes_ret = (
                    await hub.exec.boto3.client.sqs.set_queue_attributes(
                        ctx, QueueUrl=queue_url, Attributes=attributes_to_update
                    )
                )

                updated_attributes_result = updated_attributes_ret["result"]
                result["result"] = result["result"] and updated_attributes_result
                result["comment"] += (
                    hub.tool.aws.comment_utils.update_comment("aws.sqs.queue", name)
                    if updated_attributes_result
                    else updated_attributes_ret["comment"]
                )

            # tags_to_remove may contain the tags to be updated. They do not need to be deleted and then added,
            # they can be updated just by calling the hub.exec.boto3.client.sqs.tag_queue with those updated tags,
            # so they can be removed from the dict of tags to remove
            tag_keys_to_delete = list(tags_to_remove.keys() - tags_to_add.keys())
            deleted_tags = False
            if tag_keys_to_delete:
                # Delete any tags from the old state which are not in the new state
                deleted_tags_ret = await hub.exec.boto3.client.sqs.untag_queue(
                    ctx, QueueUrl=queue_url, TagKeys=tag_keys_to_delete
                )
                deleted_tags = deleted_tags_ret["result"]
                if not deleted_tags:
                    result["result"] = False
                    result["comment"] += deleted_tags_ret["comment"]

            # Update or add tags
            updated_tags = False
            if tags_to_add:
                updated_tags_ret = await hub.exec.boto3.client.sqs.tag_queue(
                    ctx, QueueUrl=queue_url, Tags=tags_to_add
                )
                updated_tags = updated_tags_ret["result"]
                if not updated_tags:
                    result["result"] = False
                    result["comment"] += updated_tags_ret["comment"]

            if updated_tags or deleted_tags:
                result["comment"] += hub.tool.aws.comment_utils.update_tags_comment(
                    tags_to_remove, tags_to_add
                )

    else:
        # Handle the case when creating the resource with --test
        if ctx.get("test", False):

            # Construct the new state in the case of creating the resource with --test
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    **present_attributes,
                    "tags": tags,
                },
            )
            result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
                "aws.sqs.queue", name
            )
        else:
            # Create a new SQS queue

            ret = await hub.exec.boto3.client.sqs.create_queue(
                ctx,
                QueueName=name,
                Attributes=plan_state_raw_attributes,
                tags=tags,
            )

            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            queue_url = ret["ret"].get("QueueUrl")
            result["comment"] += hub.tool.aws.comment_utils.create_comment(
                "aws.sqs.queue", name
            )

    if (not before) or resource_updated:
        if not ctx.get("test", False):
            # Get the new state after creating or updating the resource
            queue_state_after = await hub.exec.aws.sqs.queue.get(
                ctx,
                queue_url,
                expected_attributes=present_attributes,
                expected_tags=tags,
            )
            result["new_state"] = queue_state_after["ret"]
            result["result"] = result["result"] and queue_state_after["result"]
            result["comment"] += queue_state_after["comment"]
    else:
        # Resource already exists or there are no updates to properties
        result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
            "aws.sqs.queue", name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the queue specified by the QueueUrl, regardless of the queue's contents.  Be careful with the
    DeleteQueue action: When you delete a queue, any messages in the queue are no longer available. When you
    delete a queue, the deletion process takes up to 60 seconds. Requests you send involving that queue during the
    60 seconds might succeed. For example, a  SendMessage  request might succeed, but after 60 seconds the queue and
    the message you sent no longer exist. When you delete a queue, you must wait at least 60 seconds before creating
    a queue with the same name. Cross-account permissions don't apply to this action. For more information, see
    Grant cross-account permissions to a role and a user name in the Amazon SQS Developer Guide.

    Args:
        name(Text): The name of the SQS queue to delete.
        resource_id(Text, optional): The URL of the Amazon SQS queue to delete. Case-sensitive.
            Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
        [idem_resource_name]:
          aws.sqs.queue.absent:
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.sqs.queue.absent:
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None

    # Try to get the previous state of the resource if such is available
    if resource_id:

        # resource_id is the queue URL
        queue_state_ret = await hub.exec.aws.sqs.queue.get(ctx, resource_id)

        if queue_state_ret["result"]:
            before = queue_state_ret["ret"]
        else:
            result["result"] = False
            result["comment"] += queue_state_ret["comment"]
            return result

    if not before:
        # Resource already absent
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            "aws.sqs.queue", name
        )
    else:

        # Delete the resource

        result["old_state"] = before

        if ctx.get("test", False):
            result["comment"] += hub.tool.aws.comment_utils.would_delete_comment(
                "aws.sqs.queue", name
            )
            return result

        ret = await hub.exec.boto3.client.sqs.delete_queue(ctx, QueueUrl=resource_id)
        result["result"] = ret["result"]

        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] += hub.tool.aws.comment_utils.delete_comment(
            "aws.sqs.queue", name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**


    Returns a list of your queues in the current region.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.sqs.queue
    """

    result = {}
    queue_urls_ret = await hub.exec.boto3.client.sqs.list_queues(ctx)

    if not queue_urls_ret["result"]:
        hub.log.debug(f"Could not describe aws.sqs.queue {queue_urls_ret['comment']}")
        return {}

    # The resource_id is the queue URL
    for resource_id in queue_urls_ret["ret"]["QueueUrls"]:

        queue_state_ret = await hub.exec.aws.sqs.queue.get(ctx, resource_id)

        if not queue_state_ret["ret"]:
            hub.log.debug(queue_state_ret["comment"])
            continue

        # queue_state_ret["ret"] can contain items of type NamespaceDict
        # which cannot be represented as a valid yaml, so they need to be
        # converted to dict which is done via this copy method
        queue_state = copy.copy(queue_state_ret["ret"])
        queue_state.pop("name")
        queue_state_list = [{k: v} for k, v in queue_state.items()]

        # The queue name is the last part of the queue URL, it is used as the idem resource name in describe
        name = resource_id.split("/")[-1]
        result[name] = {"aws.sqs.queue.present": queue_state_list}

    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ce.create_cost_category_definition
hub.exec.boto3.client.ce.delete_cost_category_definition
hub.exec.boto3.client.ce.describe_cost_category_definition
hub.exec.boto3.client.ce.update_cost_category_definition
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rules: List[
        make_dataclass(
            "CostCategoryRule",
            [
                ("Value", str, field(default=None)),
                (
                    "Rule",
                    make_dataclass(
                        "Expression",
                        [
                            (
                                "Or",
                                List[
                                    make_dataclass(
                                        "Expressions",
                                        [
                                            ("Or", List[Any], field(default=None)),
                                            ("And", List[Any], field(default=None)),
                                            ("Not", Any, field(default=None)),
                                            (
                                                "Dimensions",
                                                make_dataclass(
                                                    "DimensionValues",
                                                    [
                                                        (
                                                            "Key",
                                                            str,
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "Values",
                                                            List[str],
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "MatchOptions",
                                                            List[str],
                                                            field(default=None),
                                                        ),
                                                    ],
                                                ),
                                                field(default=None),
                                            ),
                                            (
                                                "Tags",
                                                make_dataclass(
                                                    "TagValues",
                                                    [
                                                        (
                                                            "Key",
                                                            str,
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "Values",
                                                            List[str],
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "MatchOptions",
                                                            List[str],
                                                            field(default=None),
                                                        ),
                                                    ],
                                                ),
                                                field(default=None),
                                            ),
                                            (
                                                "CostCategories",
                                                make_dataclass(
                                                    "CostCategoryValues",
                                                    [
                                                        (
                                                            "Key",
                                                            str,
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "Values",
                                                            List[str],
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "MatchOptions",
                                                            List[str],
                                                            field(default=None),
                                                        ),
                                                    ],
                                                ),
                                                field(default=None),
                                            ),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("And", List[Any], field(default=None)),
                            ("Not", Any, field(default=None)),
                            ("Dimensions", Any, field(default=None)),
                            ("Tags", Any, field(default=None)),
                            ("CostCategories", Any, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "InheritedValue",
                    make_dataclass(
                        "CostCategoryInheritedValueDimension",
                        [
                            ("DimensionName", str, field(default=None)),
                            ("DimensionKey", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("Type", str, field(default=None)),
            ],
        )
    ],
    cost_category_name: str,
    rule_version: str,
    resource_id: str = None,
    split_charge_rules: List[
        make_dataclass(
            "CostCategorySplitChargeRule",
            [
                ("Source", str),
                ("Targets", List[str]),
                ("Method", str),
                (
                    "Parameters",
                    List[
                        make_dataclass(
                            "CostCategorySplitChargeRuleParameter",
                            [("Type", str), ("Values", List[str])],
                        )
                    ],
                    field(default=None),
                ),
            ],
        )
    ] = None,
    default_value: str = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new Cost Category with the requested name and rules.

    Args:
        name(Text): The unique name of the Cost Category.
        rules(List[Dict[str, Any]]): The Cost Category rules used to categorize costs. For more information, see CostCategoryRule.
            * Value (str, optional): The default value for the cost category.
            * Rule (Dict[str, Any], optional): An Expression object used to categorize costs. This supports dimensions, tags, and nested
                expressions. Currently the only dimensions supported are LINKED_ACCOUNT, SERVICE_CODE,
                RECORD_TYPE, and LINKED_ACCOUNT_NAME. Root level OR isn't supported. We recommend that you
                create a separate rule instead.  RECORD_TYPE is a dimension used for Cost Explorer APIs, and is
                also supported for Cost Category expressions. This dimension uses different terms, depending on
                whether you're using the console or API/JSON editor. For a detailed comparison, see Term
                Comparisons in the Billing and Cost Management User Guide.
                * Or (List[Dict[str, Any]], optional): Return results that match either Dimension object.
                    * Or (List[Any], optional): Return results that match either Dimension object.
                    * And (List[Any], optional): Return results that match both Dimension objects.
                    * Not (Any, optional): Return results that don't match a Dimension object.
                    * Dimensions (Dict[str, Any], optional): The specific Dimension to use for Expression.
                        * Key (str, optional): The names of the metadata types that you can use to filter and group your results. For example,
                            AZ returns a list of Availability Zones.
                        * Values (List[str], optional): The metadata values that you can use to filter and group your results. You can use
                            GetDimensionValues to find specific values.
                        * MatchOptions (List[str], optional): The match options that you can use to filter your results. MatchOptions is only applicable for
                            actions related to Cost Category. The default values for MatchOptions are EQUALS and
                            CASE_SENSITIVE.
                    * Tags (Dict[str, Any], optional): The specific Tag to use for Expression.
                        * Key (str, optional): The key for the tag.
                        * Values (List[str], optional): The specific value of the tag.
                        * MatchOptions (List[str], optional): The match options that you can use to filter your results. MatchOptions is only applicable for
                            actions related to Cost Category. The default values for MatchOptions are EQUALS and
                            CASE_SENSITIVE.
                    * CostCategories (Dict[str, Any], optional): The filter that's based on CostCategory values.
                        * Key (str, optional): The unique name of the Cost Category.
                        * Values (List[str], optional): The specific value of the Cost Category.
                        * MatchOptions (List[str], optional): The match options that you can use to filter your results. MatchOptions is only applicable for
                            actions related to cost category. The default values for MatchOptions is EQUALS and
                            CASE_SENSITIVE.
                * And (List[Any], optional): Return results that match both Dimension objects.
                * Not (Any, optional): Return results that don't match a Dimension object.
                * Dimensions (Any, optional): The specific Dimension to use for Expression.
                * Tags (Any, optional): The specific Tag to use for Expression.
                * CostCategories (Any, optional): The filter that's based on CostCategory values.
            * InheritedValue (Dict[str, Any], optional): The value the line item is categorized as if the line item contains the matched dimension.
                * DimensionName (str, optional): The name of the dimension that's used to group costs. If you specify LINKED_ACCOUNT_NAME, the
                    cost category value is based on account name. If you specify TAG, the cost category value is
                    based on the value of the specified tag key.
                * DimensionKey (str, optional): The key to extract cost category values.
            * Type (str, optional): You can define the CostCategoryRule rule type as either REGULAR or INHERITED_VALUE. The
                INHERITED_VALUE rule type adds the flexibility to define a rule that dynamically inherits the
                cost category value. This value is from the dimension value that's defined by
                CostCategoryInheritedValueDimension. For example, suppose that you want to costs to be
                dynamically grouped based on the value of a specific tag key. First, choose an inherited value
                rule type, and then choose the tag dimension and specify the tag key to use.
        cost_category_name(Text): Name of Cost Category.
        rule_version(Text): The rule schema version in this particular Cost Category.
        resource_id(Text, Optional): Cost Category ARN to identify the resource.
        split_charge_rules(List[Dict[str, Any]], optional):  The split charge rules used to allocate your charges between your Cost Category values. Defaults to None.
            * Source (str): The Cost Category value that you want to split. That value can't be used as a source or a target
            in other split charge rules. To indicate uncategorized costs, you can use an empty string as the
            source.
            * Targets (List[str]): The Cost Category values that you want to split costs across. These values can't be used as a
            source in other split charge rules.
            * Method (str): The method that's used to define how to split your source costs across your targets.
            Proportional - Allocates charges across your targets based on the proportional weighted cost of
            each target.  Fixed - Allocates charges across your targets based on your defined allocation
            percentage. >Even - Allocates costs evenly across all targets.
            * Parameters (List[Dict[str, Any]], optional): The parameters for a split charge method. This is only required for the FIXED method.
                * Type (str): The parameter type.
                * Values (List[str]): The parameter values.
        default_value(Text, Optional): The default value for the cost category.
        tags (Dict[str, Any]]): Each tag consists of a key and a value, and each key must be unique for the resource.

    Request Syntax:
        [monitor-resource-id]:
          aws.costexplorer.cost_category.present:
          - resource_id: 'string'
          - tags: 'dict'
          - cost_category_name: 'string'
          - effective_start: 'string'
          - rules:
                - Rule:
                    Dimensions:
                      Key: 'string'
                      MatchOptions:
                      - EQUALS
                      Values:
                      - 'string'
                  Type: 'string'
                  Value: 'string'
          - split_charge_rules:
                - Method: 'string'
                  Parameters:
                  - Type: 'string'
                    Values:
                    - 'string'
                    - 'string'
                  Source: 'string'
                  Targets:
                  - 'string'
              - processing_status:
                - Component: 'string'
                  Status: 'string'
              - default_value: 'string'

    Returns:
        Dict[str, str]

    Examples:

        .. code-block:: sls

            arn:aws:ce::1234567891012:costcategory/4e9662f0-5533-4fdf-8224-9bb9f82d0a39:
              aws.costexplorer.cost_category.present:
              - resource_id: arn:aws:ce::1234567891012:costcategory/4e9662f0-5533-4fdf-8224-9bb9f82d0a39
              - tags:
                  name: test_category
              - cost_category_name: test_category
              - effective_start: '2022-06-01T00:00:00Z'
              - rule_version: CostCategoryExpression.v1
              - rules:
                - Rule:
                    Dimensions:
                      Key: LINKED_ACCOUNT
                      MatchOptions:
                      - EQUALS
                      Values:
                      - '107488843946'
                  Type: REGULAR
                  Value: Alpha
                - Rule:
                    Dimensions:
                      Key: LINKED_ACCOUNT
                      MatchOptions:
                      - EQUALS
                      Values:
                      - 011922870716
                  Type: REGULAR
                  Value: Beta
                - Rule:
                    Dimensions:
                      Key: LINKED_ACCOUNT
                      MatchOptions:
                      - EQUALS
                      Values:
                      - '332986752459'
                  Type: REGULAR
                  Value: Gamma
              - split_charge_rules:
                - Method: FIXED
                  Parameters:
                  - Type: ALLOCATION_PERCENTAGES
                    Values:
                    - '40'
                    - '60'
                  Source: Alpha
                  Targets:
                  - Beta
                  - Gamma
              - processing_status:
                - Component: COST_EXPLORER
                  Status: PROCESSING
              - default_value: Other

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.boto3.client.ce.describe_cost_category_definition(
            ctx, CostCategoryArn=resource_id
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

    if before:
        result[
            "old_state"
        ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx,
            raw_resource=before["ret"]["CostCategory"],
            idem_resource_name=name,
        )
        plan_state = copy.deepcopy(result["old_state"])

        update_ret = await hub.exec.aws.costexplorer.cost_category.update_cost_category_definition(
            ctx,
            before=before["ret"]["CostCategory"],
            rules=rules,
            default_value=default_value,
            split_charge_rules=split_charge_rules,
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            plan_state = update_ret["ret"]

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_ret = await hub.exec.aws.costexplorer.cost_category.update_tags(
                ctx,
                resource_id=resource_id,
                old_tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    result["old_state"].get("tags")
                ),
                new_tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            )
            result["result"] = result["result"] and update_ret["result"]
            result["comment"] = result["comment"] + update_ret["comment"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    update_ret["ret"].get("tags")
                )

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.costexplorer.cost_category", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.costexplorer.cost_category", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cost_category_name": cost_category_name,
                    "rules": rules,
                    "split_charge_rules": split_charge_rules,
                    "default_value": default_value,
                    "tags": tags,
                    "rule_version": rule_version,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.costexplorer.cost_category", name=name
            )
            return result

        ret = await hub.exec.boto3.client.ce.create_cost_category_definition(
            ctx,
            Name=cost_category_name,
            RuleVersion=rule_version,
            Rules=rules,
            DefaultValue=default_value,
            SplitChargeRules=split_charge_rules,
            ResourceTags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
        resource_id = ret["ret"]["CostCategoryArn"]

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.ce.describe_cost_category_definition(
                ctx, CostCategoryArn=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result[
                "new_state"
            ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
                ctx,
                raw_resource=after["ret"]["CostCategory"],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """
    **Autogenerated function**

    Deletes a cost category.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, Optional): Cost Category ARN to identify the resource. Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.costexplorer.cost_category.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
        return result

    before = await hub.exec.boto3.client.ce.describe_cost_category_definition(
        ctx, CostCategoryArn=resource_id
    )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx, raw_resource=before["ret"]["CostCategory"], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx, raw_resource=before["ret"]["CostCategory"], idem_resource_name=name
        )
        ret = await hub.exec.boto3.client.ce.delete_cost_category_definition(
            ctx, CostCategoryArn=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of aws cost categories.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.costexplorer.cost_category
    """
    result = {}
    ret = await hub.exec.boto3.client.ce.list_cost_category_definitions(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not list Cost Category Definitions {ret['comment']}")
        return {}

    for costcategory in ret["ret"]["CostCategoryReferences"]:
        resource_id = costcategory.get("CostCategoryArn")
        ret = await hub.exec.boto3.client.ce.describe_cost_category_definition(
            ctx, CostCategoryArn=resource_id
        )

        if not ret["result"]:
            hub.log.debug(
                f"Could not describe Cost Category Definitions {ret['comment']}"
            )
            return {}

        resource_translated = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx,
            raw_resource=ret["ret"].get("CostCategory"),
            idem_resource_name=resource_id,
        )
        result[resource_id] = {
            "aws.costexplorer.cost_category.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.attach_internet_gateway
hub.exec.boto3.client.ec2.create_internet_gateway
hub.exec.boto3.client.ec2.delete_internet_gateway
hub.exec.boto3.client.ec2.describe_internet_gateways
hub.exec.boto3.client.ec2.detach_internet_gateway
resource = hub.tool.boto3.resource.create(ctx, "ec2", "InternetGateway", name)
hub.tool.boto3.resource.exec(resource, attach_to_vpc, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": [
            "aws.ec2.elastic_ip.absent",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    vpc_id: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an internet gateway for use with a VPC. After creating the internet gateway, you attach it to a VPC
    using AttachInternetGateway. For more information about your VPC and internet gateway, see the Amazon Virtual
    Private Cloud User Guide.

    Args:
        name(Text): An Idem name to identify the internet gateway resource.
        resource_id(Text, optional): AWS Internet Gateway ID.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the internet gateway resource. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        vpc_id(List, optional): This list can contain only single element. This is ID of VPC to which internet gateway attaches.
        If vpc_id is not passed, then attach/detach operation is ignored.
        If user passes empty list, then attached vpc is detached.

    Request Syntax:
        [internet_gateway-name]:
          aws.ec2.internet_gateway.present:
          - vpc_id:
            - 'string'
          - tags:
            - Key: 'string'
              Value: 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test-igw:
             aws.ec2.internet_gateway.present:
             - vpc_id:
               - vpc-d04fbf46
             - tags:
               - Key: Name
                 Value: test-igw
    """

    plan_state = {}
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    # Skip fetching resource if there is no change
    is_state_updated = True
    before = None
    if resource_id:
        before = await hub.exec.boto3.client.ec2.describe_internet_gateways(
            ctx=ctx, DryRun=False, InternetGatewayIds=[resource_id]
        )

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before and before["result"]:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_internet_gateway_to_present(
            resource=before["ret"]["InternetGateways"][0], idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])
        result["comment"] = (f"'{name}' already exists",)
        try:
            if tags is not None and tags != result["old_state"].get("tags"):
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx=ctx,
                    resource_id=resource_id,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                result["result"] = update_ret["result"]
                result["comment"] = result["comment"] + update_ret["comment"]
                if ctx.get("test", False) and update_ret["result"]:
                    plan_state["tags"] = update_ret["ret"]
            ret = await hub.exec.aws.ec2.internet_gateway.update_attachments(
                ctx=ctx,
                internet_gateway_name=resource_id,
                vpc_id=vpc_id,
                before=result["old_state"],
            )
            result["result"] = ret["result"]
            result["comment"] = result["comment"] + ret["comment"]
            if tags is None and vpc_id is None:
                is_state_updated = False
            if ctx.get("test", False) and ret["ret"].get("updated_vpc_id") is not None:
                plan_state["vpc_id"] = ret["ret"]["updated_vpc_id"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "vpc_id": vpc_id,
                    "tags": tags,
                    "resource_id": name,
                },
            )
            result["comment"] = (
                f"Would create aws.ec2.internet_gateway '{name}' and attach to vpc '{vpc_id[0]}'",
            )
            return result
        try:
            ret = await hub.exec.boto3.client.ec2.create_internet_gateway(
                ctx,
                TagSpecifications=[
                    {
                        "ResourceType": "internet-gateway",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = ret["ret"]["InternetGateway"]["InternetGatewayId"]
            result["comment"] = (f"Created '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
        if vpc_id:
            try:
                ret = await hub.exec.boto3.client.ec2.attach_internet_gateway(
                    ctx, InternetGatewayId=resource_id, VpcId=vpc_id[0]
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = result["comment"] + (
                    f"Created '{name}' and attached to vpc '{vpc_id[0]}'",
                )
            except hub.tool.boto3.exception.ClientError as e:
                result["result"] = False
                result["comment"] = result["comment"] + (
                    f"Created '{name}'. Attach to '{vpc_id[0]}' failed with error: {e}",
                )
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif is_state_updated or not (before and before["result"]):
            after = await hub.exec.boto3.client.ec2.describe_internet_gateways(
                ctx, DryRun=False, InternetGatewayIds=[resource_id]
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_internet_gateway_to_present(
                resource=after["ret"].get("InternetGateways")[0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified internet gateway. If resource is attached to VPC. then it is detached first and then deleted.

    Args:
        name(Text): An Idem name to identify the internet gateway resource.
        resource_id(Text, optional): AWS Internet Gateway ID.

    Request Syntax:
        [internet_gateway-name]:
          aws.ec2.internet_gateway.absent:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

           igw-test:
            aws.ec2.internet_gateway.absent:
              - name: igw-test
              - resource_id: '12345678'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.internet_gateway", name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(
        ctx, "ec2", "InternetGateway", resource_id
    )
    before = await hub.tool.boto3.resource.describe(resource)
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.internet_gateway", name=name
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_internet_gateway_to_present(
            resource=before, idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = (f"Would delete aws.ec2.internet_gateway '{name}'",)
            return result
        else:
            try:
                gateway_resource = before
                if gateway_resource["Attachments"]:
                    resource.detach_from_vpc(
                        VpcId=gateway_resource["Attachments"][0]["VpcId"]
                    )
                ret = await hub.exec.boto3.client.ec2.delete_internet_gateway(
                    ctx, InternetGatewayId=resource_id
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = (f"Deleted '{name}'",)
            except hub.tool.boto3.exception.ClientError as e:
                result["result"] = False
                result["comment"] = (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Describes one or more of your internet gateways.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.internet_gateway
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_internet_gateways(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe internet_gateway {ret['comment']}")
        return {}

    for internet_gateway in ret["ret"]["InternetGateways"]:
        resource_id = internet_gateway.get("InternetGatewayId")
        resource_converted = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_internet_gateway_to_present(
                resource=internet_gateway, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.internet_gateway.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_converted.items()
            ]
        }
    return result



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub, ctx, name: str, role_name: str, policy_arn: str, resource_id: str = None
) -> Dict[str, Any]:
    """
    Attaches the specified managed policy to the specified IAM role. When you attach a managed policy to a role,
     the managed policy becomes part of the role's permission (access) policy.

    Args:
        hub:
        ctx:
        name(Text): A name to represent the operation. This name is only for logging purpose. It is not used to
         attach a policy to a role.
        role_name(Text): The name (friendly name, not ARN) of the role to attach a policy.
         This parameter allows (through its regex pattern ) a string of characters consisting of upper and lowercase
          alphanumeric characters with no spaces. You can also include any of the following characters: _+=,.@-
        policy_arn(Text): The Amazon Resource Name (ARN) of the IAM policy you want to attach.
        resource_id(Text, Optional): The identifier for this object

    Request Syntax:
        [iam-attach-role-policy-name]:
          aws.iam.role_policy_attachment.present:
          - resource_id: 'string'
          - role_name: 'string'
          - policy_arn: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-policy-temp-name:
              aws.iam.role_policy_attachment.present:
                - role_name: idem-test-role-name
                - policy_arn: arn:aws:iam::aws:policy/ReadOnlyAccess

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    try:
        before_ret = (
            await hub.exec.aws.iam.role_policy_attachment.is_role_policy_attached(
                ctx, role_name=role_name, policy_arn=policy_arn
            )
        )
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = (f"{e.__class__.__name__}: {e}",)
        result["result"] = False
        return result

    # If before_ret has False result and has an error message in comment, then immediately return with the error,
    # and if role does not exists - NoSuchEntityException is expected
    if before_ret["result"] is False and before_ret["comment"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if before_ret["result"]:
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_attachment_to_present(
            role_name, policy_arn
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (f"aws.iam.role_policy_attachment '{name}' already exists",)
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "role_name": role_name,
                        "policy_arn": policy_arn,
                    },
                )
                result["comment"] = (
                    f"Would create aws.iam.role_policy_attachment. Role {role_name}, policy {policy_arn}",
                )
                return result

            ret = await hub.exec.boto3.client.iam.attach_role_policy(
                ctx,
                RoleName=role_name,
                PolicyArn=policy_arn,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = (f"Created aws.iam.role_policy_attachment '{name}'",)
            result[
                "new_state"
            ] = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_attachment_to_present(
                role_name, policy_arn
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, role_name: str, policy_arn: str
) -> Dict[str, Any]:
    """

    Removes the specified managed policy from the specified role.

    Args:
        name(Text): The name of the AWS IAM role policy.
        role_name(Text): The name (friendly name, not ARN) of the role to attach a policy.
         This parameter allows (through its regex pattern ) a string of characters consisting of upper and lowercase
          alphanumeric characters with no spaces. You can also include any of the following characters: _+=,.@-
        policy_arn(Text): The Amazon Resource Name (ARN) of the IAM policy you want to attach.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.iam.role_policy_attachment.absent:
                - name: value
                - role_name: value
                - policy_arn: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    try:
        before_ret = (
            await hub.exec.aws.iam.role_policy_attachment.is_role_policy_attached(
                ctx, role_name=role_name, policy_arn=policy_arn
            )
        )
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = (f"{e.__class__.__name__}: {e}",)
        result["result"] = False
        return result

    if before_ret["result"] is False and before_ret["comment"]:
        if "NoSuchEntityException" in before_ret["comment"][0]:
            result["comment"] = (
                f"aws.iam.role_policy_attachment '{name}' already absent",
            )
        else:
            result["result"] = False
            result["comment"] = before_ret["comment"]
        return result

    if not before_ret["result"]:
        result["comment"] = (f"aws.iam.role_policy_attachment '{name}' already absent",)
    else:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_attachment_to_present(
                role_name, policy_arn
            )
            if ctx.get("test", False):
                result["comment"] = (
                    f"Would detach aws.iam.role_policy_attachment. Role '{role_name}', policy '{policy_arn}'",
                )
                return result

            ret = await hub.exec.boto3.client.iam.detach_role_policy(
                ctx, RoleName=role_name, PolicyArn=policy_arn
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (
                f"Detached aws.iam.role_policy_attachment. Role '{role_name}', policy '{policy_arn}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the names of the attached managed policies of all IAM roles. If there are no managed policies attached
     to the specified role, the operation returns an empty dict.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.role_policy_attachment

    """

    result = {}
    # To describe all the attached role policies of all the roles, we first need to list all the roles, then get all the
    # attached policies
    ret_roles = await hub.exec.boto3.client.iam.list_roles(ctx)
    if not ret_roles["result"]:
        hub.log.debug(f"Could not describe role {ret_roles['comment']}")
        return {}
    role_name_list = [role.get("RoleName") for role in ret_roles["ret"]["Roles"]]
    for role_name in role_name_list:
        ret_attached_policies = (
            await hub.exec.boto3.client.iam.list_attached_role_policies(
                ctx=ctx, RoleName=role_name
            )
        )
        if not ret_attached_policies["result"]:
            hub.log.warning(
                f"Could not get attached policy list with role {role_name} with error"
                f" {ret_attached_policies['comment']} . Describe will skip this role and continue."
            )
        else:
            resources = ret_attached_policies["ret"].get("AttachedPolicies")
            for resource in resources:
                translated_resource = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_attachment_to_present(
                    role_name, resource.get("PolicyArn")
                )
                resource_id = f"{role_name}-{translated_resource['policy_arn']}"
                result[resource_id] = {
                    "aws.iam.role_policy_attachment.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in translated_resource.items()
                    ]
                }

    return result



from typing import Any
from typing import Dict
from typing import List

from dict_tools import data


async def update(
    hub,
    ctx,
    current_state: Dict[str, Any],
    description: str,
    regular_expression_list: List[Dict[str, str]],
) -> Dict[str, Any]:

    # Description and regular_expression_list can be updated. Check if they are modified and update
    result = dict(ret=None, comment=(), result=True)

    if current_state.get("description") != description or data.recursive_diff(
        current_state.get("regular_expression_list"),
        regular_expression_list,
        ignore_order=True,
    ):
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.wafv2.update_regex_pattern_set(
                ctx=ctx,
                Name=current_state.get("name"),
                Scope=current_state.get("scope"),
                Id=current_state.get("resource_id"),
                Description=description,
                RegularExpressionList=regular_expression_list,
                LockToken=current_state.get("LockToken"),
            )
            result["result"] = update_ret["result"]
            if not result["result"]:
                result["comment"] = update_ret["comment"]
                return result
        result["ret"] = {
            "description": description,
            "regular_expression_list": regular_expression_list,
        }
    return result


def check_if_invalid_scope_and_region(hub, region, scope):
    if region != "us-east-1" and scope == "CLOUDFRONT":
        return "CLOUDFRONT scope is available only in US East (N. Virginia) Region, us-east-1"
    return None
